#!/usr/bin/env python3
import re
import collections
import os
import sys
import struct
import argparse

from PyQt5 import QtCore, QtGui, QtWidgets

# if started via invoking this file directly (like would happen with source sitting on disk),
# insert our relative msgtools root dir into the sys.path, so *our* msgtools is used, not
# any other already in the path.
if __name__ == '__main__':
    srcroot=os.path.abspath(os.path.dirname(os.path.abspath(__file__))+"/../..")
    sys.path.insert(1, srcroot)
from msgtools.lib.messaging import Messaging
import msgtools.lib.gui
import msgtools.lib.createmsgdialog
import msgtools.lib.msgjson as msgjson

DESCRIPTION='''DebugPrint provides a graphical interface that allows you to view debug messages, and send
    binary messages and text commands.  It tries to be pretty flexible about what messages are used, by
    looking up messages by MessageAlias, and by accepting Thread/Stream as synonyms.'''

class DebugStream(QtWidgets.QWidget):
    messageOutput = QtCore.pyqtSignal(object)
    statusUpdate = QtCore.pyqtSignal(str)
    def __init__(self, streamID, debugDevice, debugWidget):
        super(DebugStream, self).__init__()
        self.debugDevice = debugDevice
        self.debugWidget = debugWidget
        self.name = "Stream" + str(streamID)
        self.widget = msgtools.lib.gui.TreeWidget()
        self.widget.debugStream = self
        self.streamID = streamID
        self.widget.priority = 1000
        self.widget.itemDoubleClicked.connect(self.debugWidget.tableDataDoubleClicked)

        if self.debugWidget.setThresholdMsg:
            self.m_tabMenu = QtWidgets.QMenu()
            self.m_tabMenu.addSection("Device " + self.debugDevice.route + ", Stream " + str(streamID))
            if self.debugWidget.getDeviceInfoMsg:
                queryAction = self.m_tabMenu.addAction("Query Device")
                queryAction.triggered.connect(self.debugDevice.getDeviceInfo)
            if self.debugWidget.getStreamInfoMsg:
                queryAction = self.m_tabMenu.addAction("Query Stream")
                queryAction.triggered.connect(self.getStreamInfo)
            self.m_tabMenu.addSection("")
            actionGroup = QtWidgets.QActionGroup(self)
            for debugLevel in self.debugWidget.setThresholdMsg.Priorities.keys():
                debugThresholdChangedAction = self.m_tabMenu.addAction(debugLevel)
                debugThresholdChangedAction.setCheckable(1)
                actionGroup.addAction(debugThresholdChangedAction)
                debugThresholdChangedAction.triggered.connect(self.debugThresholdChanged)

        # add it to the tab widget, so the user can see it
        added = 0
        tabName = debugDevice.name+", " + self.name
        for index in range(0,self.debugWidget.tabWidget.count()):
            if(tabName < self.debugWidget.tabWidget.tabText(index)):
                self.debugWidget.tabWidget.insertTab(index, self.widget, tabName)
                added = 1
                break
        if not added:
            self.debugWidget.tabWidget.addTab(self.widget, tabName)

        # add table header, one column for each message field
        tableHeader = []
        tableHeader.append("Time (ms)")
        tableHeader.append("Priority")
        tableHeader.append("File")
        tableHeader.append("Line")
        tableHeader.append("Message")
        
        self.widget.setHeaderLabels(tableHeader)

        # daisy-chain our signals to the debugDevice's signals
        self.messageOutput.connect(self.debugDevice.messageOutput)
        self.statusUpdate.connect(self.debugDevice.statusUpdate)

    def clear(self):
        self.widget.clear()

    def Rename(self, deviceName, streamName):
        self.name = streamName
        newName = deviceName + ", " + streamName
        tabIndex = self.debugWidget.tabWidget.indexOf(self.widget)
        if tabIndex < 0:
            print("Warning!  Couldn't find debug tab for device " + route + " stream " + str(streamID))
        else:
            self.debugWidget.tabWidget.setTabText(tabIndex, newName)

    def debugThresholdChanged(self):
        action = self.sender()
        msg = self.debugWidget.setThresholdMsg()
        Messaging.SetMsgRoute(msg, self.debugDevice.route.split("/"))
        msg.SetStreamID(self.streamID)
        msg.SetDebugThreshold(action.text())
        self.messageOutput.emit(msg)

    def getStreamInfo(self):
        msg = self.debugWidget.getStreamInfoMsg()
        Messaging.SetMsgRoute(msg, self.debugDevice.route.split("/"))
        try:
            msg.SetStreamID(self.streamID)
        except AttributeError:
            msg.SetThreadID(self.streamID)
        self.messageOutput.emit(msg)

    def AddPrintMessage(self, msg, firstTime):
        msgStringList = []
        try:
            msgStringList.append(str(msg.hdr.GetTime()))
        except AttributeError:
            msgStringList.append("Unknown")
        text = ""
        filename = ""
        linenumber = ""
        try:
            priority = msg.GetPriority()
            priorityInt = msg.GetPriority(1)
            priorityIntError = msg.Priorities["Error"]
        except AttributeError:
            priority = ""
            priorityInt = 0
            priorityIntError = 255
        if type(msg) == self.debugWidget.printfMsg:
            text = msg.GetBuffer()
            matchObj = re.search( r'(.*), (\d+): (.*)', text)
            if matchObj != None:
                filename = matchObj.group(1).strip()
                linenumber = matchObj.group(2).strip()
                text = matchObj.group(3).strip()
        elif type(msg) == self.debugWidget.printfIDMsg:
            # need to look up the format string in the dictionary, and use it to print a string to display!
            formatStringId = msg.GetFormatStringID()
            strFormatError = 0
            if formatStringId < len(self.debugDevice.dictionary):
                info = self.debugDevice.dictionary[formatStringId]
                filename = info.filename
                linenumber = info.linenumber
                # need to evaluate the formatStr and parameters to produce a new string!
                try:
                    paramsNeeded = info.formatStr.count("%")
                    params = []
                    for i in range(0,min(paramsNeeded,4)):
                        try:
                            value = int(msg.GetParameters(i))
                        except struct.error:
                            value = 0
                        params.append(value)
                    text = info.formatStr % tuple(params)
                except TypeError:
                    strFormatError = 1
            else:
                strFormatError = 1

            # if we couldn't format the string for whatever reason, fall back to just displaying the parameters
            if strFormatError:
                text = str(formatStringId) +"("
                for i in range(0,4):
                    try:
                        text += str(msg.GetParameters(i))+","
                    except struct.error:
                        pass
                if text.endswith(","):
                    text = text[:-1]
                text += ")"
                    
        msgStringList.append(str(priority))
        msgStringList.append(filename)
        msgStringList.append(linenumber)
        msgStringList.append(text)

        msgItem = QtWidgets.QTreeWidgetItem(None,msgStringList)
        color = 0
        if(priority == "Error" or priorityInt > priorityIntError):
            color = QtCore.Qt.red
        elif(priority == "Warning"):
            color = QtGui.QColor("darkOrange")
        elif(priority != "Info"):
            color = QtCore.Qt.blue
        if color != 0:
            font = msgItem.font(0)
            brush = msgItem.foreground(0)
            font.setBold(1)
            brush.setColor(color)
            for column in range(0,5):
                msgItem.setFont(column, font)
                msgItem.setForeground(column, brush)
                msgItem.setBackground(column, brush)
        # if new priority is more severe than old priority, change color
        oldPriority = self.widget.priority
        if priorityInt < oldPriority:
            index = self.debugWidget.tabWidget.indexOf(self.widget)
            if color == 0:
                color = QtCore.Qt.darkGreen
            self.debugWidget.tabWidget.tabBar().setTabTextColor(index, color)
            self.widget.priority = priorityInt
        
        self.widget.addTopLevelItem(msgItem)
        if(self.debugWidget.autoscroll):
            self.widget.scrollToItem(msgItem)
        # we should have an option to log PrintfID messages as text, or else we should
        # send them out as Printf messages (as long as they don't go to the device!),
        # so standard logs will show them in an easy to read form

        if firstTime:
            for col in range(0, 5):
                self.widget.resizeColumnToContents(col)

class DebugDevice(QtWidgets.QWidget):
    messageOutput = QtCore.pyqtSignal(object)
    statusUpdate = QtCore.pyqtSignal(str)
    def __init__(self, debugWidget, route, dictionaryFilename):
        super(DebugDevice, self).__init__()
        self.debugWidget = debugWidget
        self.route = route
        self.name = "Dev"+route
        self.dictionary = []
        self.dictionaryID = None
        self.dictionaryFilename = None
        if dictionaryFilename:
            print("reading dictionary["+route+"] " + dictionaryFilename)
            self.ReadDictionary(dictionaryFilename)

        self.streams = {}
        
        # daisy-chain our signals to the debugWidget's signals
        self.messageOutput.connect(self.debugWidget.messageOutput)
        self.statusUpdate.connect(self.debugWidget.statusUpdate)

    def clear(self):
        for streamID, stream in self.streams.items():
            stream.clear()

    def getDeviceInfo(self):
        msg = self.debugWidget.getDeviceInfoMsg()
        Messaging.SetMsgRoute(msg, self.route.split("/"))
        self.messageOutput.emit(msg)

    def ReadDictionary(self, filename):
        self.dictionaryFilename = filename
        self.dictionary = []
        nextId = 0
        with open(filename, 'r') as formatStringFile:
            lines = formatStringFile.read().splitlines()
            for line in lines:
                matchObj = re.search( r'(\d+).*:\s*"([^"]*)", (.*), (\d+)', line)
                if matchObj != None:
                    DebugInfo = collections.namedtuple('DebugInfo', ['formatStr', 'filename', 'linenumber'])
                    id = matchObj.group(1).strip()
                    if int(id) != nextId:
                        print("ERROR! Format string ID " + str(id) + " != " + str(nextId))
                    formatStr = matchObj.group(2).strip()
                    filename = matchObj.group(3).strip()
                    linenumber = matchObj.group(4).strip()
                    info = DebugInfo(formatStr, filename, linenumber)
                    self.dictionary.append(info)
                    nextId += 1
                else:
                    matchObj = re.search( r'Dictionary md5 is (.*)', line)
                    if matchObj != None:
                        self.dictionaryID = matchObj.group(1).strip()
                        print("Read dictionary " + self.dictionaryID)

    def ProcessMessage(self, msg):
        if type(msg) == self.debugWidget.deviceInfoMsg:
            dictionaryID = ""
            for i in range(msg.GetDebugStringDictionaryID.count):
                dictionaryID += '{:02x}'.format(msg.GetDebugStringDictionaryID(i))
            if dictionaryID != self.dictionaryID:
                #TODO If dictionary ID changed or wasn't previously set, load the dictionary.
                s = "Device %s needs to load dictionary %s" % (self.name, dictionaryID)
                print(s)
                self.statusUpdate.emit(s)

            try:
                deviceName = msg.GetDeviceName()
            except AttributeError:
                deviceName = msg.GetName()
            self.name = deviceName
            for streamID,stream in self.streams.items():
                stream.Rename(deviceName, stream.name)
            return

        if type(msg) == self.debugWidget.streamInfoMsg:
            try:
                streamID = msg.GetStreamID()
            except AttributeError:
                streamID = msg.GetThreadID()
            try:
                streamName = msg.GetStreamName()
            except AttributeError:
                try:
                    streamName = msg.GetThreadName()
                except AttributeError:
                    streamName = msg.GetName()
            if not streamID in self.streams:
                self.streams[streamID] = DebugStream(streamID, self, self.debugWidget)

            self.streams[streamID].Rename(self.name, streamName)
            return

        # only handle Printf and PrintfID messages!
        try:
            if type(msg) != self.debugWidget.printfMsg and type(msg) != self.debugWidget.printfIDMsg:
                return
        except NameError:
            return
        
        try:
            streamID = msg.GetStreamID()
        except AttributeError:
            try:
                streamID = msg.GetThreadID()
            except AttributeError:
                streamID = 0
        
        if streamID in self.streams:
            firstTime = 0
        else:
            firstTime = 1
            self.streams[streamID] = DebugStream(streamID, self, self.debugWidget)

        self.streams[streamID].AddPrintMessage(msg, firstTime)

# this is a widget used for debug purposes.
# it incorporates:
# 1) entry text box
#   a. to send binary messages
#   b. to send text commands
# 2) cmd/response text display to show:
#   a. entered commands
#   b. responses
# 3) separate tabs for debug text output
# 4) controls to set threshold for debug output per stream of each device
class MsgDebugWidget(QtWidgets.QWidget):
    messageOutput = QtCore.pyqtSignal(object)
    autocompleted   = QtCore.pyqtSignal(str)
    statusUpdate = QtCore.pyqtSignal(str)
    def __init__(self, debugdicts, parent=None):
        super(MsgDebugWidget, self).__init__()
        
        # find classes for print messages.
        try:
            self.printfMsg = Messaging.Messages.DebugPrintf
        except AttributeError:
            self.printf = None
        try:
            self.printfIDMsg = Messaging.Messages.DebugPrintfID
        except AttributeError:
            self.printfIDMsg = None
        try:
            self.getDeviceInfoMsg = Messaging.Messages.GetDeviceDebugInfo
        except AttributeError:
            self.getDeviceInfoMsg = None
        try:
            self.getStreamInfoMsg = Messaging.Messages.GetStreamDebugInfo
        except AttributeError:
            self.getStreamInfoMsg = None
        try:
            self.setThresholdMsg = Messaging.Messages.SetDebugThreshold
        except AttributeError:
            self.setThresholdMsg = None
        try:
            self.deviceInfoMsg = Messaging.Messages.DebugDeviceInfo
        except AttributeError:
            self.deviceInfoMsg = None
        try:
            self.streamInfoMsg = Messaging.Messages.DebugStreamInfo
        except AttributeError:
            self.streamInfoMsg = None
        
        # tab widget to show multiple stream of print statements, one per tab
        self.tabWidget = QtWidgets.QTabWidget()
        self.tabWidget.currentChanged.connect(self.tabChanged)
        self.tabWidget.tabBar().tabBarClicked.connect(self.createTabContextMenu)
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(self.tabWidget)
        self.setLayout(vbox)
        
        # tab for text commands
        self.textEntryWidget = msgtools.lib.gui.MsgCommandWidget(self)
        self.tabWidget.addTab(self.textEntryWidget, "Cmd/Reply")
        self.textEntryWidget.commandEntered.connect(self.newCommandEntered)
        self.textEntryWidget.messageEntered.connect(self.newMessageEntered)
        self.textEntryWidget.autocompleted.connect(self.autocompleted)
        
        # tracking what reply to expect
        self.expectedReply = None
        
        self.fileWatcher = QtCore.QFileSystemWatcher()

        # hash table of devices by route
        self.devices = {}
        
        for d in debugdicts:
            route, dictName = d.split("=")
            self.devices[route] = DebugDevice(self, route, dictName)
            self.fileWatcher.addPath(dictName)

        # whether we should autoscroll as data is added
        self.autoscroll = 1
        self.scrollAction = QtWidgets.QAction('&Scroll', self)
        self.scrollAction.setCheckable(1)
        self.scrollAction.setChecked(self.autoscroll)
        self.scrollAction.triggered.connect(self.switchScroll)

        self.fileWatcher.fileChanged.connect(self.fileChanged)

    def clearTab(self):
        self.tabWidget.currentWidget().clear()
    
    def clearAllTabs(self):
        for route, device in self.devices.items():
            device.clear()
        
    def switchScroll(self):
        self.autoscroll = not self.autoscroll
        self.scrollAction.setChecked(self.autoscroll)

    def tableDataDoubleClicked(self, treeWidgetItem, column):
        self.autoscroll = not self.autoscroll
        self.scrollAction.setChecked(self.autoscroll)

    def createTabContextMenu(self, tabIndex):
        if QtWidgets.QApplication.mouseButtons() != QtCore.Qt.RightButton:
            return
        try:
            self.tabWidget.currentWidget().debugStream
        except AttributeError:
            return
        self.tabWidget.currentWidget().debugStream.m_tabMenu.popup(QtGui.QCursor.pos())

    def newCommandEntered(self, cmd):
        try:
            tc = Messaging.Messages.MsgText.Command()
            tc.SetBuffer(cmd)
            tc.hdr.SetDataLength(len(cmd)+1)
            self.messageOutput.emit(tc)
        except:
            self.textEntryWidget.addText("\nWARNING: No message named/aliased to MsgText.Command exists, not sending text\n> ")
    
    def newMessageEntered(self, msg):
        self.expectedReply = msg.MsgName().rsplit(".",1)[0]
        self.messageOutput.emit(msg)

    def ReadDictionary(self, route, filename):
        nextId = 0
        if route in self.devices:
            self.devices[route].ReadDictionary(filename)
        else:
            self.devices[route] = DebugDevice(self, route, filename)
    
    def ProcessMessage(self, msg):
        # Create the DebugDevice if it doesn't already exist
        route = "/".join(Messaging.MsgRoute(msg))
        if not route in self.devices:
            self.devices[route] = DebugDevice(self, route, None)

        # handle text replies
        alreadyPrintedText = False
        try:
            if type(msg) == Messaging.Messages.MsgText.Response:
                text = msg.GetBuffer()
                text = text.replace("\\n","\n")
                text = text.replace("\\r","")
                self.textEntryWidget.addText(text)
                if "error" in text.lower():
                    color = QtCore.Qt.red
                elif "warning" in text.lower():
                    color = QtGui.QColor("darkOrange")
                else:
                    color = QtGui.QColor("darkGreen")
                self.tabWidget.tabBar().setTabTextColor(0, color)
                alreadyPrintedText = True
        except:
            pass
        
        if not alreadyPrintedText and self.expectedReply and msg.MsgName().startswith(self.expectedReply):
            outputString = msgjson.toJson(msg)
            self.tabWidget.tabBar().setTabTextColor(0, QtGui.QColor("darkGreen"))
            self.textEntryWidget.addText(outputString+"\n> ")

        self.devices[route].ProcessMessage(msg)

    def tabChanged(self, index):
        # when user selects a tab, change it's color back to black
        w = self.tabWidget.widget(index)
        w.priority = 1000
        self.tabWidget.tabBar().setTabTextColor(index, QtCore.Qt.black)

    def fileChanged(self, path):
        for route, device in self.devices.items():
            if device.dictionaryFilename == path:
                device.ReadDictionary(device.dictionaryFilename)
                print("dictionary " + path + " found at device " + route)
                self.statusUpdate.emit("updating device " + route)
                QtCore.QTimer.singleShot(3000, self.clearStatus)
                return
        print("dictionary " + path + " not found")
    def clearStatus(self):
        self.statusUpdate.emit("")

class DebugPrint(msgtools.lib.gui.Gui):
    def __init__(self, args, parent=None):
        msgtools.lib.gui.Gui.__init__(self, "Debug Print 0.1", args, parent)

        if args.debugdicts:
            debugdictList = args.debugdicts.split(",")
        else:
            debugdictList = []
        self.debugWidget = MsgDebugWidget(debugdictList)
        self.debugWidget.statusUpdate.connect(self.statusUpdate)
        self.setCentralWidget(self.debugWidget)
        self.debugWidget.messageOutput.connect(self.SendMsg)
        
        # menu items to change view
        clearAction = QtWidgets.QAction('&Clear', self)
        clearAllAction = QtWidgets.QAction('Clear &All', self)

        menubar = self.menuBar()
        viewMenu = menubar.addMenu('&View')
        viewMenu.addAction(clearAction)
        viewMenu.addAction(clearAllAction)
        viewMenu.addAction(self.debugWidget.scrollAction)

        clearAction.triggered.connect(self.debugWidget.clearTab)
        clearAllAction.triggered.connect(self.debugWidget.clearAllTabs)

        # menu items to control the device output
        if self.debugWidget.getDeviceInfoMsg or self.debugWidget.setThresholdMsg:
            controlMenu = menubar.addMenu('&Control')

        if self.debugWidget.getDeviceInfoMsg:
            queryDeviceAction = QtWidgets.QAction('&Query Devices', self)
            queryDeviceAction.triggered.connect(self.QueryDeviceDebugInfo)
            controlMenu.addAction(queryDeviceAction)
        
        if self.debugWidget.getStreamInfoMsg:
            queryStreamAction = QtWidgets.QAction('&Query Streams', self)
            queryStreamAction.triggered.connect(self.QueryStreamDebugInfo)
            controlMenu.addAction(queryStreamAction)

        if self.debugWidget.setThresholdMsg:
            setThresholdAction = QtWidgets.QAction('&Set Threshold', self)
            setThresholdAction.triggered.connect(self.SetThresholdDialog)
            controlMenu.addAction(setThresholdAction)

        # event-based way of getting messages
        self.RxMsg.connect(self.debugWidget.ProcessMessage)

    def QueryDeviceDebugInfo(self):
        # for each device we've seen any message from, query the dictionary
        for route, device in self.debugWidget.devices.items():
            # should we re-query devices we already have a dictionary ID for?
            if 1: #device.dictionaryID == None:
                device.getDeviceInfo()

    def QueryStreamDebugInfo(self):
        # for each stream of each device we've seen any message from, query
        for route, device in self.debugWidget.devices.items():
            for streamID, stream in device.streams.items():
                # should we re-query streams we already have a name for?
                if 1: #stream.name == None:
                    stream.getStreamInfo()

    def SetThresholdDialog(self):
        # open a dialog to let use set debug output threshold
        # for a specified device and stream
        dlg = msgtools.lib.createmsgdialog.CreateMsgDialog(self, self.debugWidget.setThresholdMsg)
        if dlg.exec_():
            self.SendMsg(dlg.msg)

def main():
    # Setup a command line processor...
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser = msgtools.lib.gui.Gui.addBaseArguments(parser)
    parser.add_argument('--debugdicts', help=''''Dictionaries to use to lookup debug message format strings, of the form 0=pathToDictionary0,1=pathToDictionary1.''')
    args = parser.parse_args()

    app = QtWidgets.QApplication(sys.argv)
    msgApp = DebugPrint(args)
    msgApp.show()
    sys.exit(app.exec_())

# main starts here
if __name__ == '__main__':
    main()



from PyQt5 import QtCore, QtWidgets
from PyQt5.QtCore import QObject

from msgtools.lib.messaging import Messaging
from msgtools.lib.header_translator import HeaderTranslator
from BluetoothHeader import BluetoothHeader

BTHS = BluetoothHeader.SIZE

# We require bluez, available on Windows and Linux
import bluetooth
import threading
import select
import time

class BluetoothRFCOMMConnection(QObject):
    disconnected = QtCore.pyqtSignal(object)
    messagereceived = QtCore.pyqtSignal(object)
    statusUpdate = QtCore.pyqtSignal(str)

    def __init__(self, deviceBTAddr, deviceBTPort=None):
        super(BluetoothRFCOMMConnection, self).__init__(None)

        self.removeClient = QtWidgets.QPushButton("Remove")
        self.removeClient.pressed.connect(self.onDisconnected)

        self.statusLabel = QtWidgets.QLabel()
        self.subscriptions = {}
        self.subMask = 0
        self.subValue = 0
        self.isHardwareLink = True
        self.basetime = time.time()
        
        if deviceBTPort is None:
            services = bluetooth.find_service(address=deviceBTAddr,
                                              uuid='00001101-0000-1000-8000-00805F9B34FB')
            if len(services)<=0:
                deviceBTPort = 8
            else:
                deviceBTPort = services[0]['port']
        
        self.socket = bluetooth.BluetoothSocket(bluetooth.RFCOMM)
        self.socket.connect((deviceBTAddr, deviceBTPort))
        #self.socket.disconnected.connect(self.onDisconnected)

        self.btsock_buffer = b''
        self.btsock_outgoing = b''

        self.hdrTranslator = HeaderTranslator(BluetoothHeader, Messaging.hdr)
        
        self.name = "Bluetooth RFCOMM " + deviceBTAddr
        self.statusLabel.setText(self.name)

        self.thread = threading.Thread(target=self.rfcommThread)
        self.thread.daemon = True
        self.thread.start()

    def start(self):
        pass

    def widget(self, index):
        if index == 0:
            return self.removeClient
        if index == 1:
            return self.statusLabel
        return None

    def rfcommThread(self):
        while True:
            ret = select.select([self.socket],
                                [self.socket],
                                [],
                                0.1)
            if len(ret[2]) > 0:
                #self.statusUpdate.emit("some message?")
                self.socket.close()
                return

            if len(ret[0]) > 0:
                # we've got data to read
                self.btsock_buffer += self.socket.recv(4096)

            if len(self.btsock_outgoing)>0 and len(ret[1])>0:
                sent = self.socket.send(self.btsock_outgoing)
                self.btsock_outgoing = self.btsock_outgoing[sent:]

            while len(self.btsock_buffer) >= BTHS:
                hdr = BluetoothHeader(self.btsock_buffer[0:BTHS])
                total_len = BTHS + hdr.GetDataLength()
                if len(self.btsock_buffer) >= total_len:
                    # we've got enough data for the whole message
                    hdr = BluetoothHeader(self.btsock_buffer)
                    self.btsock_buffer = self.btsock_buffer[total_len:]
                    
                    networkMsg = self.hdrTranslator.translate(hdr)
                    
                    self.messagereceived.emit(networkMsg)
                else:
                    # we don't quite have enough data for a whole message
                    break

    def onDisconnected(self):
        print("self.disconnected.emit(self)")
        self.socket.close()
        self.disconnected.emit(self)
        print("self.disconnected.emit(self)")

    def sendMsg(self, networkMsg):
        btMsg = self.hdrTranslator.translate(networkMsg)
        # if we can't translate the message, just return
        if btMsg == None:
            return
        self.btsock_outgoing += btMsg.rawBuffer().raw

    def stop(self):
        pass

def PluginConnection(param=""):
    btArgs = param.split(",")
    if len(btArgs)>1:
        btArgs[1] = int(btArgs[1])
    return BluetoothRFCOMMConnection(*btArgs)

def PluginEnabled():
    return True

import collections
PluginInfo = collections.namedtuple('PluginInfo', ['name', 'enabled', 'connect_function'])
plugin_info = PluginInfo('Bluetooth RF Comm', PluginEnabled, PluginConnection)



import msgtools.parser.parser as MsgParser
from msgtools.parser.MsgUtils import *

def fieldType(field):
    fieldTypeDict = \
    {"uint64":"error", "uint32":"long","uint16": "int",   "uint8": "short",
     "int64":"long",   "int32":"int",  "int16": "short",  "int8": "byte",
      "float64":"double", "float32":"float"}
    typeStr = field["Type"]
    return fieldTypeDict[typeStr]

def fieldAccessorType(field):
    fieldTypeDict = \
    {"uint64":"long", "uint32":"int", "uint16": "short", "uint8": "",
     "int64":"long",   "int32":"int",  "int16": "short",  "int8": "",
      "float64":"double", "float32":"float"}
    typeStr = field["Type"]
    type = fieldTypeDict[typeStr]
    return type.capitalize()

def fieldCastType(field):
    fieldTypeDict = \
    {"uint64":"long", "uint32":"int", "uint16": "short", "uint8": "byte",
     "int64":"long",   "int32":"int",  "int16": "short",  "int8": "byte",
      "float64":"double", "float32":"float"}
    typeStr = field["Type"]
    type = fieldTypeDict[typeStr]
    return type.capitalize()

def fieldPromotionFn(field):
    fieldTypeDict = \
    {"uint64":"error", "uint32":"FieldAccess.toUnsignedLong", "uint16": "FieldAccess.toUnsignedInt", "uint8": "FieldAccess.toUnsignedInt"}
    typeStr = field["Type"]
    return fieldTypeDict[typeStr]

def fnHdr(field):
    ret = "// %s %s, (%s to %s)" % (MsgParser.fieldDescription(field), MsgParser.fieldUnits(field), MsgParser.fieldMin(field), MsgParser.fieldMax(field))
    return ret

def typeForScaledInt(field):
    numBits = MsgParser.fieldNumBits(field)
    if numBits > 24:
        return "double"
    return "float"

def getFn(field):
    loc = str(MsgParser.fieldLocation(field))
    param = ""
    if MsgParser.fieldCount(field) > 1:
        loc += "+idx*" + str(MsgParser.fieldArrayElementOffset(field))
        param += "int idx"
    access = "m_data.get%s(%s)" % (fieldAccessorType(field), loc)
    if field["Type"].startswith("u"):
        access = fieldPromotionFn(field)+"("+access+")"
    access = "("+fieldType(field)+")"+access
    access = getMath(access, field, "("+typeForScaledInt(field)+")", 'f')
    retType = fieldType(field)
    if "Offset" in field or "Scale" in field:
        retType = typeForScaledInt(field)
    #elif "Enum" in field:
    #    retType = field["Enum"]
    #    access = retType + ".construct(" + access + ")"
    ret = '''\
%s
public %s Get%s(%s)
{
    return %s;
}''' % (fnHdr(field), retType, field["Name"], param, access)
    return ret

def setFn(field):
    paramType = fieldType(field)
    valueString = setMath("value", field, '('+fieldType(field)+')', 'f')
    if "Offset" in field or "Scale" in field:
        paramType = typeForScaledInt(field)
    #elif "Enum" in field:
    #    valueString = valueString + ".intValue()"
    #    paramType = field["Enum"]
    param = paramType + " value"
    loc = str(MsgParser.fieldLocation(field))
    if MsgParser.fieldCount(field) > 1:
        loc += "+idx*" + str(MsgParser.fieldArrayElementOffset(field))
        param += ", int idx"
    ret = '''\
%s
public void Set%s(%s)
{
    m_data.put%s(%s, (%s)%s);
}''' % (fnHdr(field), field["Name"], param, fieldAccessorType(field), loc, fieldCastType(field).lower(), valueString)
    return ret

def getBitsFn(field, bits, bitOffset, numBits):
    access = "(Get%s() >> %s) & %s" % (field["Name"], str(bitOffset), MsgParser.Mask(numBits))
    access = getMath(access, bits, "("+typeForScaledInt(bits)+")", 'f')
    retType = fieldType(field)
    if "Offset" in bits or "Scale" in bits:
        retType = typeForScaledInt(bits)
    #elif "Enum" in bits:
    #    retType = bits["Enum"]
    #    access = retType + ".construct(" + access + ")"
    else:
        access = "("+retType+")" + "(" + access + ")"
    ret = '''\
%s
public %s Get%s()
{
    return %s;
}''' % (fnHdr(bits), retType, MsgParser.BitfieldName(field, bits), access)
    return ret

def setBitsFn(field, bits, bitOffset, numBits):
    paramType = fieldType(field)
    intType = paramType
    valueString = setMath("value", bits, '('+fieldType(field)+')', 'f')
    if "Offset" in bits or "Scale" in bits:
        paramType = typeForScaledInt(bits)
    #elif "Enum" in bits:
    #    valueString = valueString + ".intValue()"
    #    paramType = bits["Enum"]
    ret = '''\
%s
public void Set%s(%s value)
{
    Set%s((%s)((Get%s() & ~(%s << %s)) | ((%s & %s) << %s)));
}''' % (fnHdr(bits), MsgParser.BitfieldName(field, bits), paramType, field["Name"], intType, field["Name"], MsgParser.Mask(numBits), str(bitOffset), valueString, MsgParser.Mask(numBits), str(bitOffset))
    return ret

def accessors(msg):
    gets = []
    sets = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            gets.append(getFn(field))
            sets.append(setFn(field))
            bitOffset = 0
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    gets.append(getBitsFn(field, bits, bitOffset, numBits))
                    sets.append(setBitsFn(field, bits, bitOffset, numBits))
                    bitOffset += numBits

    return gets+sets

def languageConst(value):
    ret = value
    if isinstance(ret, str) and ret.isdigit():
        ret = int(ret)
    if isinstance(ret, int):
        if(ret > 2**31):
            ret = str(ret)+"L"
        else:
            ret = str(ret)
    elif not isinstance(ret, str):
        ret = str(ret)
    return ret

def initField(field):
    type = fieldType(field)
    if "Default" in field:
        if MsgParser.fieldCount(field) > 1:
            ret = "for (int i=0; i<" + str(MsgParser.fieldCount(field)) + "; i++)\n"
            ret += "    Set%s((%s)%s, i);" % (field["Name"], type, languageConst(field["Default"])) 
            return ret;
        else:
            return  "Set%s((%s)%s);" % (field["Name"],type, languageConst(field["Default"]))
    return ""

def initBitfield(field, bits):
    type = fieldType(field)
    if "Default" in bits:
        return  "Set%s((%s)%s);" % (MsgParser.BitfieldName(field, bits), type, str(bits["Default"]))
    return ""

def initCode(msg):
    ret = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            fieldInit = initField(field)
            if fieldInit:
                ret.append(fieldInit)
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    bits = initBitfield(field, bits)
                    if bits:
                        ret.append(bits)

    return ret

def optionValue(option):
    ret = int(option["Value"])
    if ret > 2**31:
        return str(ret)+"L"
    return str(ret)

def enums(e):
    ret = ""
    for enum in e:
        ret +=  "public enum " + enum["Name"]+" {\n"
        ret += "    "
        for option in enum["Options"]:
            ret += OptionName(option)+"("+str(optionValue(option)) + '), '
        ret = ret[:-2]
        ret += ";\n"
        ret += '''\
    private final long id;
    {0}(long id) {{ this.id = id; }}
    static Map<Long, {0}> map = new HashMap<>();
    static {{
        for ({0} key : {0}.values()) {{
            map.put(key.id, key);
        }}
    }}
    public long intValue() {{ return id; }}
    public static {0} construct(long value) {{ return map.get(value); }}
}}\n'''.format(enum["Name"])
    return ret

def fieldReflection(field):
    loc = str(MsgParser.fieldLocation(field))
    params = "FieldInfo";
    params += "("
    params += '"'+field["Name"] + '"'
    params += ', "' + MsgParser.fieldDescription(field) + '"'
    params += ', "' + MsgParser.fieldUnits(field) + '"'
    params += ", " + str(MsgParser.fieldCount(field))
    params += ")"
    return params

def fieldBitsReflection(field, bits, bitOffset, numBits):
    loc = str(MsgParser.fieldLocation(field))
    params = "FieldInfo";
    params += "("
    params += '"'+bits["Name"] + '"'
    params += ', "' + MsgParser.fieldDescription(bits) + '"'
    params += ', "' + MsgParser.fieldUnits(bits) + '"'
    params += ", " + str(MsgParser.fieldCount(bits))
    params += ")"
    return params

def reflection(msg):
    ret = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            ret.append(fieldReflection(field))
            bitOffset = 0
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    ret.append(fieldBitsReflection(field, bits, bitOffset, numBits))
                    bitOffset += numBits

    return "\n".join(ret)

def fieldMin(field):
    val = MsgParser.fieldMin(field)
    ret = str(val)
    if ret == 'DBL_MIN':
        ret = 'Double.MIN_VALUE'
    elif ret == 'FLT_MIN':
        ret = 'Float.MIN_VALUE'
    else:
        try:
            if val > fieldStorageMax("uint32"):
                ret += 'L'
        except TypeError:
            pass
    return ret

def fieldMax(field):
    val = MsgParser.fieldMax(field)
    ret = str(val)
    if ret == 'DBL_MAX':
        ret = 'Double.MAX_VALUE'
    elif ret == 'FLT_MAX':
        ret = 'Float.MAX_VALUE'
    else:
        try:
            if val > fieldStorageMax("int32"):
                ret += 'L'
        except TypeError:
            pass
    return ret

def genericInfo(field, loc, type):
    params  = '    public static final int loc = ' + loc + ';\n'
    params += '    public static final %s max = (%s)%s;\n' % (type, type, fieldMax(field))
    params += '    public static final %s min = (%s)%s;\n' % (type, type, fieldMin(field))
    params += '    public static final String units = "' + str(MsgParser.fieldUnits(field)) + '"' + ';\n'
    params += '    public static final int count = ' + str(MsgParser.fieldCount(field)) + ';\n'
    if "Default" in field:
        params += '    public static final %s defaultValue = (%s)%s;\n' % (type, type, languageConst(field["Default"]))
    if "Scale" in field:
        params += '    public static final float scale = (float)' + str(field["Scale"]) + ';\n'
    if "Offset" in field:
        params += '    public static final float offset = (float)' + str(field["Offset"]) + ';\n'
    return params
    
def fieldInfo(field):
    retType = fieldType(field)
    if "Offset" in field or "Scale" in field:
        retType = typeForScaledInt(field)

    params  = 'public class ' + field["Name"] + 'FieldInfo {\n'
    params += genericInfo(field, str(MsgParser.fieldLocation(field)), retType)
    params += '};\n'
    return params

def fieldBitsInfo(field, bits, bitOffset, numBits):
    retType = fieldType(field)
    if "Offset" in bits or "Scale" in bits:
        retType = typeForScaledInt(bits)

    params  = 'public class ' + bits["Name"] + 'FieldInfo {\n'
    params += genericInfo(bits, str(MsgParser.fieldLocation(field)), retType)
    params += '    public static final int bitOffset = ' + str(bitOffset) + ';\n'
    params += '    public static final int numBits   = ' + str(numBits) + ';\n'
    params += '};\n'
    return params

def fieldInfos(msg):
    ret = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            ret.append(fieldInfo(field))
            bitOffset = 0
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    ret.append(fieldBitsInfo(field, bits, bitOffset, numBits))
                    bitOffset += numBits

    return "\n".join(ret)

def declarations(msg):
    return []

def getMsgID(msg):
    ret = ""
    if "Fields" in msg:
        for field in msg["Fields"]:
            if "IDBits" in field:
                numBits = field["IDBits"]
                param = ""
                getStr = "Get"+field["Name"]+"("+param+")"
                #if "Enum" in field:
                #    getStr = getStr+".intValue()"
                ret =  addShift(ret, getStr, numBits)
            if "Bitfields" in field:
                for bitfield in field["Bitfields"]:
                    if "IDBits" in bitfield:
                        numBits = bitfield["IDBits"]
                        param = ""
                        getStr = "Get"+BitfieldName(field, bitfield)+"("+param+")"
                        #if "Enum" in bitfield:
                        #    getStr = getStr+".intValue()"
                        ret =  addShift(ret, getStr, numBits)
    return ret
    
def setMsgID(msg):
    ret = ""
    numBits = 0
    if "Fields" in msg:
        for field in reversed(msg["Fields"]):
            type = fieldType(field)
            if "IDBits" in field:
                if numBits != 0:
                    ret += "\nid = id >> " + str(numBits)+"\n"
                numBits = field["IDBits"]
                setStr = "id & "+Mask(numBits)
                #if "Enum" in field:
                #    setStr = field["Enum"]+".construct("+setStr+")"
                #else:
                setStr = "("+type+")("+setStr+")"
                ret +=  "Set"+field["Name"]+"("+setStr+")"
            if "Bitfields" in field:
                for bitfield in reversed(field["Bitfields"]):
                    if "IDBits" in bitfield:
                        if numBits != 0:
                            ret += "\nid = id >> " + str(numBits)+"\n"
                        numBits = bitfield["IDBits"]
                        setStr = "id & "+Mask(numBits)
                        #if "Enum" in bitfield:
                        #    setStr = bitfield["Enum"]+".construct("+setStr+")"
                        #else:
                        setStr = "("+type+")("+setStr+")"
                        ret +=  "Set"+bitfield["Name"]+"("+setStr+")"
    return ret

oneOutputFilePerMsg = True



#    <OUTPUTFILENAME>
#    Created <DATE> from:
#        Messages = <INPUTFILENAME>
#        Template = <TEMPLATEFILENAME>
#        Language = <LANGUAGEFILENAME>
#
#                     AUTOGENERATED FILE, DO NOT EDIT
import struct
from collections import OrderedDict
from msgtools.lib.conversions import *
from msgtools.lib.messaging import *
import msgtools.lib.messaging as msg
from msgtools.lib.message import Message

class <MSGNAME>(Message):
    ID = <MSGID>
    SIZE = <MSGSIZE>
    MSG_OFFSET = Messaging.hdrSize
    # Enumerations
    <ENUMERATIONS>
    
    def __init__(self, messageBuffer=None, **kwargs):
        super(<MSGNAME>, self).__init__(messageBuffer, id=<MSGNAME>.ID, size=<MSGNAME>.SIZE)
        doInit = 0
        if messageBuffer == None:
            self.initialize()
        self.set_fields(**kwargs)

    def initialize(self):
            <INIT_CODE>
            pass

    def rawBuffer(self):
        # this is a trick to get us to store a copy of a pointer to a buffer, rather than making a copy of the buffer
        return self.msg_buffer_wrapper["msg_buffer"]

    @staticmethod
    def MsgName():
        return "<MSGDESCRIPTOR>"
    # Accessors
    <ACCESSORS>

    # Reflection information
    fields = [ \
        <REFLECTION>\
    ]

Messaging.Register("<MSGDESCRIPTOR>", <MSGNAME>.ID, <MSGNAME>)
Messaging.AddAlias("<MSGALIAS>", <MSGNAME>.ID, <MSGNAME>)



import sys
import datetime
import inspect
import argparse

from PyQt5 import QtGui, QtWidgets, QtCore, QtNetwork

from .app import *

import msgtools.lib.msgcsv as msgcsv

# tree widget item with support for sorting by a column
class TreeWidgetItem(QtWidgets.QTreeWidgetItem):
    def __init__(self, parent, stringList):
        QtWidgets.QTreeWidgetItem.__init__(self, parent, stringList)

    def __lt__(self, otherItem):
        column = self.treeWidget().sortColumn()
        try:
            return float(self.text(column)) < float(otherItem.text(column))
        except ValueError:
            return self.text(column) < otherItem.text(column)

# tree widget that we can copy-and-paste on
class TreeWidget(QtWidgets.QTreeWidget):
    def __init__(self):
        super(TreeWidget, self).__init__() # what constructor parameters?!
        self.setSelectionMode(self.ContiguousSelection)
        self.setSelectionBehavior(self.SelectRows)

    def keyPressEvent(self, keyEvent):
        if keyEvent.matches(QtGui.QKeySequence.Copy):
            self.copySelection()
        else:
            super(TreeWidget, self).keyPressEvent(keyEvent)

    def copySelection(self):
        selectedItems = self.selectedItems()
        copiedText = ""
        h = self.headerItem()
        for col in range(h.columnCount()):
            if col != 0:
                copiedText += ", "
            copiedText += h.text(col)
        copiedText += "\n"
        
        for itemNumber in range(len(selectedItems)):
            w = selectedItems[itemNumber]
            for col in range(w.columnCount()):
                # treat first column as time, convert HH:MM:SS.sss to a float
                if col == 0:
                    t = w.text(col)
                    if ":" in t:
                        time_components = t.split(':')
                        t = int(time_components[0]) * 3600 + int(time_components[1]) * 60 + float(time_components[2])
                        copiedText += str(t)
                    else:
                        copiedText += str(t)
                else:
                    copiedText += ", "
                    copiedText += w.text(col)
            copiedText += "\n"
        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setText(copiedText)

# text widget that allows us to push msg data directly into the text fields
class MsgTextWidget(QtWidgets.QWidget):
    def __init__(self, msgClass, allowedFields=[]):
        super(MsgTextWidget, self).__init__()
        self.grid = QtWidgets.QGridLayout()
        self.setLayout(self.grid)
        from msgtools.lib.unknownmsg import UnknownMsg
        if msgClass == UnknownMsg:
            self.showHeader = True
        else:
            self.showHeader = False
        self.allowedFields = dict((fieldName,True) for fieldName in allowedFields)
        tableHeader = []
        if self.fieldAllowed("Time"):
            timeUnits = Messaging.findFieldInfo(Messaging.hdr.fields, "Time").units
            if timeUnits == "ms":
                timeUnits = "s"
            tableHeader.append("Time ("+timeUnits+")")
        if self.showHeader:
            for fieldInfo in Messaging.hdr.fields:
                if len(fieldInfo.bitfieldInfo) == 0:
                    if self.fieldAllowed(fieldInfo.name):
                        tableHeader.append(fieldInfo.name)
                else:
                    for bitInfo in fieldInfo.bitfieldInfo:
                        if self.fieldAllowed(bitInfo.name):
                            tableHeader.append(bitInfo.name)
        for fieldInfo in msgClass.fields:
            if self.fieldAllowed(fieldInfo.name):
                tableHeader.append(fieldInfo.name)
            for bitInfo in fieldInfo.bitfieldInfo:
                if self.fieldAllowed(bitInfo.name):
                    tableHeader.append(bitInfo.name)
        self.setHeaderLabels(tableHeader)
    
    def fieldAllowed(self, fieldName):
        if len(self.allowedFields) == 0:
            return True
        return fieldName in self.allowedFields

    def setHeaderLabels(self, headerLabels):
        i = 0
        for h in headerLabels:
            self.grid.addWidget(QtWidgets.QLabel(h), 0, i)
            self.grid.addWidget(QtWidgets.QLabel(''), 1, i)
            i += 1
    
    def setDataLabels(self, dataLabels, dataAlerts):
        i = 0
        for t in dataLabels:
            self.grid.itemAtPosition(1,i).widget().setText(t)
            if dataAlerts[i]:
                #TODO Change text color
                pass
            i += 1

    def addData(self, msg, autoscroll=1):
        msgStringList = []
        columnAlerts = []
        if self.fieldAllowed("Time"):
            try:
                timeVal = msg.hdr.GetTime()
                timeInfo = Messaging.findFieldInfo(msg.hdr.fields, "Time")
                if timeInfo.units == "ms":
                    timeVal = timeVal / 1000.0
                timeVal = datetime.datetime.fromtimestamp(timeVal, datetime.timezone.utc)
            except AttributeError:
                timeVal = datetime.datetime.now()
            timeVal = timeVal.strftime('%H:%M:%S.%f')[:-3]
            msgStringList.append(timeVal)
            columnAlerts.append(0)
        if self.showHeader:
            for fieldInfo in Messaging.hdr.fields:
                if len(fieldInfo.bitfieldInfo) == 0:
                    if self.fieldAllowed(fieldInfo.name):
                        fieldValue = str(Messaging.get(msg.hdr, fieldInfo))
                        msgStringList.append(fieldValue)
                else:
                    for bitInfo in fieldInfo.bitfieldInfo:
                        if self.fieldAllowed(bitInfo.name):
                            fieldValue = str(Messaging.get(msg.hdr, bitInfo))
                            msgStringList.append(fieldValue)
        for fieldInfo in type(msg).fields:
            if(fieldInfo.count == 1):
                if self.fieldAllowed(fieldInfo.name):
                    fieldValue = str(Messaging.get(msg, fieldInfo))
                    msgStringList.append(fieldValue)
                    columnAlerts.append(Messaging.getAlert(msg, fieldInfo))
                for bitInfo in fieldInfo.bitfieldInfo:
                    if self.fieldAllowed(bitInfo.name):
                        fieldValue = str(Messaging.get(msg, bitInfo))
                        msgStringList.append(fieldValue)
                        columnAlerts.append(Messaging.getAlert(msg, bitInfo))
            else:
                if self.fieldAllowed(fieldInfo.name):
                    columnText = ""
                    alert = 0
                    for i in range(0,fieldInfo.count):
                        fieldValue = Messaging.get(msg, fieldInfo, i)
                        # if the value is what is given when we go off the end of an array, break.
                        if fieldInfo.type == "int" and fieldValue == "UNALLOCATED":
                            break
                        columnText += str(fieldValue)
                        if Messaging.getAlert(msg, fieldInfo, i):
                            alert = 1
                        if(i<fieldInfo.count-1):
                            columnText += ", "
                    msgStringList.append(columnText)
                    columnAlerts.append(alert)

        self.setDataLabels(msgStringList, columnAlerts)

# tree widget that allows us to push msg data directly into the tree
class MsgTreeWidget(TreeWidget):
    MAX_ROWS = 1000
    ROWS_TO_DELETE = 50
    def __init__(self, msgClass, keyField=None, maxRows=1000, rowsToDelete=50):
        super(MsgTreeWidget, self).__init__()
        from msgtools.lib.unknownmsg import UnknownMsg
        if msgClass == UnknownMsg:
            self.showHeader = True
        else:
            self.showHeader = False
        # add table header, one column for each message field
        tableHeader = []
        timeUnits = Messaging.findFieldInfo(Messaging.hdr.fields, "Time").units
        if timeUnits == "ms":
            timeUnits = "s"
        tableHeader.append("Time ("+timeUnits+")")
        if self.showHeader:
            for fieldInfo in Messaging.hdr.fields:
                if len(fieldInfo.bitfieldInfo) == 0:
                    tableHeader.append(fieldInfo.name)
                else:
                    for bitInfo in fieldInfo.bitfieldInfo:
                        tableHeader.append(bitInfo.name)
        for fieldInfo in msgClass.fields:
            tableHeader.append(fieldInfo.name)
            for bitInfo in fieldInfo.bitfieldInfo:
                tableHeader.append(bitInfo.name)
        self.setHeaderLabels(tableHeader)
        # configure the header so we can click on it to sort
        self.header().setSectionsClickable(1)
        self.header().setSortIndicatorShown(1)
        # show sort indicator ascending on Time, if not sorting, because we append incoming messages
        self.header().setSortIndicator(0, QtCore.Qt.AscendingOrder)
        self.header().sectionClicked.connect(self.tableHeaderClicked)
        
        self.maxRows = maxRows
        if rowsToDelete > maxRows:
            rowsToDelete = maxRows
        self.rowsToDelete = rowsToDelete

        # key field for our message, that we display one row per value of
        self.keyField = keyField

        self.firstTimeDataAdded = 1

    def addData(self, msg, autoscroll=1):
        msgStringList = []
        columnAlerts = []
        try:
            timeVal = msg.hdr.GetTime()
            timeInfo = Messaging.findFieldInfo(msg.hdr.fields, "Time")
            if timeInfo.units == "ms":
                timeVal = timeVal / 1000.0
            timeVal = datetime.datetime.fromtimestamp(timeVal, datetime.timezone.utc)
        except AttributeError:
            timeVal = datetime.datetime.now()
        timeVal = timeVal.strftime('%H:%M:%S.%f')[:-3]
        msgStringList.append(timeVal)
        columnAlerts.append(0)
        keyColumn = -1
        columnCounter = 1
        if self.showHeader:
            for fieldInfo in Messaging.hdr.fields:
                if len(fieldInfo.bitfieldInfo) == 0:
                    fieldValue = str(Messaging.get(msg.hdr, fieldInfo))
                    msgStringList.append(fieldValue)
                    columnCounter += 1
                else:
                    for bitInfo in fieldInfo.bitfieldInfo:
                        fieldValue = str(Messaging.get(msg.hdr, bitInfo))
                        msgStringList.append(fieldValue)
                        columnCounter += 1
        for fieldInfo in type(msg).fields:
            if(fieldInfo.count == 1):
                fieldValue = str(Messaging.get(msg, fieldInfo))
                msgStringList.append(fieldValue)
                columnAlerts.append(Messaging.getAlert(msg, fieldInfo))
                if fieldInfo.name == self.keyField:
                    keyValue = fieldValue
                    keyColumn = columnCounter
                columnCounter += 1
                for bitInfo in fieldInfo.bitfieldInfo:
                    fieldValue = str(Messaging.get(msg, bitInfo))
                    msgStringList.append(fieldValue)
                    columnAlerts.append(Messaging.getAlert(msg, bitInfo))
                    if bitInfo.name == self.keyField:
                        keyValue = fieldValue
                        keyColumn = columnCounter
                    columnCounter += 1
            else:
                columnText = ""
                alert = 0
                for i in range(0,fieldInfo.count):
                    fieldValue = Messaging.get(msg, fieldInfo, i)
                    # if the value is what is given when we go off the end of an array, break.
                    if fieldInfo.type == "int" and fieldValue == "UNALLOCATED":
                        break
                    columnText += str(fieldValue)
                    if Messaging.getAlert(msg, fieldInfo, i):
                        alert = 1
                    if(i<fieldInfo.count-1):
                        columnText += ", "
                msgStringList.append(columnText)
                columnAlerts.append(alert)
                columnCounter += 1

        msgItem = TreeWidgetItem(None,msgStringList)
        for column in range(0, len(columnAlerts)):
            if columnAlerts[column]:
                font = msgItem.font(column)
                brush = msgItem.foreground(column)
                font.setBold(1)
                brush.setColor(QtCore.Qt.red)
                msgItem.setFont(column, font)
                msgItem.setForeground(column, brush)
                msgItem.setBackground(column, brush)
        if self.keyField != None and keyColumn >= 0:
            # find row that has key field that matches ours
            foundAndReplaced = 0
            for i in range(0, self.topLevelItemCount()):
                item = self.topLevelItem(i)
                if item.text(keyColumn) == keyValue:
                    foundAndReplaced = 1
                    self.takeTopLevelItem(i)
                    self.insertTopLevelItem(i, msgItem)
            if not foundAndReplaced:
                self.addTopLevelItem(msgItem)
                self.sortItems(keyColumn, QtCore.Qt.AscendingOrder)
        else:
            self.addTopLevelItem(msgItem)
            if(autoscroll):
                self.scrollToItem(msgItem)
            # when deleting, make a bunch of room so that if we're auto scrolling we have a bit of time before it
            # shifts the data.  Otherwise the user can't read stuff if it's going by too fast.
            if self.topLevelItemCount() > self.maxRows:
                for i in range(0, self.rowsToDelete):
                    self.takeTopLevelItem(0)
                    
        if self.firstTimeDataAdded:
            count = 0
            for fieldInfo in type(msg).fields:
                self.resizeColumnToContents(count)
                count += 1
        self.firstTimeDataAdded = 0
        
    def tableHeaderClicked(self, column):
        fieldName = self.headerItem().text(column)
        if self.keyField == None or self.keyField != fieldName:
            self.keyField = fieldName
            self.header().setSortIndicator(column, QtCore.Qt.AscendingOrder)
            self.sortItems(column, QtCore.Qt.AscendingOrder)
            valueToRemove = None
            for i in range(self.topLevelItemCount()-1, -1 ,-1):
                item = self.topLevelItem(i)
                if not valueToRemove == None and item.text(column) == valueToRemove:
                    self.takeTopLevelItem(i)
                else:
                    valueToRemove = item.text(column)
        else:
            self.keyField = None
            self.header().setSortIndicator(0, QtCore.Qt.AscendingOrder)

class LineEditWithHistory(QtWidgets.QLineEdit):
    tabPressed = QtCore.pyqtSignal()
    def __init__(self, parent=None):
        super(LineEditWithHistory, self).__init__()
        self.commandHistory = []
        self.placeInHistory = 0

    def setToHistoryItem(self, index):
        self.placeInHistory = index
        if self.placeInHistory > len(self.commandHistory) - 1:
            self.placeInHistory = len(self.commandHistory)
            self.setText('')
        else:
            if self.placeInHistory < 0:
                self.placeInHistory = 0
            try:
                self.setText(self.commandHistory[self.placeInHistory])
            except IndexError:
                self.setText('')
    
    # disable tab focus so we get tab keys delivered via keyPressEvent.
    def focusNextPrevChild(self, next):
        return False
    
    def addToHistory(self, lineOfText):
        if len(self.commandHistory) == 0 or self.commandHistory[-1] != lineOfText:
            self.commandHistory.append(lineOfText)
            if len(self.commandHistory) > 20:
                self.commandHistory.pop(0)
        self.placeInHistory = len(self.commandHistory)

    def keyPressEvent(self, keyEvent):
        if keyEvent.key() == QtCore.Qt.Key_Return:
            # add to history
            self.addToHistory(self.text())
        elif keyEvent.key() == QtCore.Qt.Key_Up:
            #up in history
            self.setToHistoryItem(self.placeInHistory - 1)
        elif keyEvent.key() == QtCore.Qt.Key_Down:
            # down in history
            self.setToHistoryItem(self.placeInHistory + 1)
        elif keyEvent.key() == QtCore.Qt.Key_Tab:
            self.tabPressed.emit()
            return
        super(LineEditWithHistory, self).keyPressEvent(keyEvent)

class MsgCommandWidget(QtWidgets.QWidget):
    commandEntered = QtCore.pyqtSignal(str)
    messageEntered = QtCore.pyqtSignal(object)
    autocompleted   = QtCore.pyqtSignal(str)
    def __init__(self, parent=None):
        super(MsgCommandWidget, self).__init__()
        self.textBox = QtWidgets.QPlainTextEdit()
        self.textBox.setReadOnly(True)
        self.lastHtml = False
        self.lineEdit = LineEditWithHistory()
        self.lineEdit.returnPressed.connect(self.returnPressed)
        self.lineEdit.tabPressed.connect(self.tabPressed)
        
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(self.textBox)
        vbox.addWidget(self.lineEdit)
        self.setLayout(vbox)
    
    def tabPressed(self):
        lineOfText = self.lineEdit.text()
        autocomplete, help = msgcsv.csvHelp(lineOfText)
        if autocomplete:
            self.lineEdit.setText(autocomplete)
            self.autocompleted.emit(autocomplete)
        if help:
            self.addText('\n'+help+'\n>\n')

    def returnPressed(self):
        lineOfText = self.lineEdit.text()
        self.addText(lineOfText)
        msg = msgcsv.csvToMsg(lineOfText)
        if msg:
            self.messageEntered.emit(msg)
            self.addText(" -> Msg\n")
        else:
            self.commandEntered.emit(lineOfText)
        self.lineEdit.setText("")

    def addText(self, text, errorKnown=0):
        lines = text.splitlines()
        for i in range(len(lines)):
            line = lines[i]
            if i < len(lines)-1:
                line = line + '\n'
            elif i == len(lines)-1 and text.endswith('\n'):
                line = line + '\n'
            self.addLine(line, errorKnown)
    
    def addLine(self, line, errorKnown):
        if errorKnown == 2 or 'error' in line.lower() or 'fail' in line.lower():
            self.textBox.appendHtml('<font color="red">'+line+'</font>')
            self.lastHtml = True
        elif errorKnown == 1 or 'warning' in line.lower():
            self.textBox.appendHtml('<font color="orange">'+line+'</font>')
            self.lastHtml = True
        else:
            if self.lastHtml:
                self.textBox.appendHtml('<font color="black"> </font>')
            self.textBox.moveCursor (QtGui.QTextCursor.End)
            self.textBox.insertPlainText(line)
            self.textBox.moveCursor (QtGui.QTextCursor.End)
            self.lastHtml = False

    def clear(self):
        self.textBox.clear()
    
    def addToHistory(self, lineOfText):
        self.lineEdit.addToHistory(lineOfText)
    
    def saveState(self):
        ret = ""
        for h in self.lineEdit.commandHistory:
            ret = ret + "|" + h
        return ret
    
    def restoreState(self, state):
        if state:
            history = state.split("|")
            for h in history:
                self.lineEdit.addToHistory(h)

class Gui(App, QtWidgets.QMainWindow):
    @classmethod
    def addBaseArguments(cls, parser):
        '''
        Adds base app arguments to the provided ArgParser

        returns the parser
        '''

        # Just delegate - this design choice is two fold - it prevents
        # code using the Gui from including the App as a dependency.
        # It also allows us to inject Gui specific arguments into this
        # base class later.
        return App.addBaseArguments(parser)

    '''Gui base class that provides standard connection and status UI

        name - the name of the app - will be displayed in the title bar of the window
        args - argparse.Namespace representing parsed arguments.  See app.py for
               the standard options provided as part of this Gui/App framework used
               in most MsgTools utilities.
        parent - a parent Qt Widget to attach to
    '''
    def __init__(self, name, args, parent=None):

        QtWidgets.QMainWindow.__init__(self,parent)
        App.__init__(self, name, args)

        # make a status bar to print status messages
        self.status = QtWidgets.QLabel("Initializing")
        self.statusBar().addPermanentWidget(self.status)
        # hook it up to our base class statusUpdate signal
        self.statusUpdate.connect(self.status.setText)
        
        # a checkbox for connection state to the status bar
        self.connectionCheckbox = QtWidgets.QCheckBox("")
        self.statusBar().addPermanentWidget(self.connectionCheckbox)
        # hook it up to connect/disconnect
        self.connectionCheckbox.clicked.connect(self.connectionCheckedChanged)
        self.connectionChanged.connect(self.connectionChangedSlot)

        QtCore.QTimer.singleShot(0, self.delayedInit)
        self.setWindowTitle(self.name)

        # create menu items, connect them to socket operations
        if(self.connectionType.lower() == "socket" or self.connectionType.lower() == "qtsocket"):
            connectAction = QtWidgets.QAction('&Connect', self)
            disconnectAction = QtWidgets.QAction('&Disconnect', self)

            menubar = self.menuBar()
            connectMenu = menubar.addMenu('&Connect')
            connectMenu.addAction(connectAction)
            connectMenu.addAction(disconnectAction)
            connectAction.triggered.connect(self.chooseHost)
            disconnectAction.triggered.connect(self.CloseConnection)
    
    def delayedInit(self):
        if hasattr(self.__class__, 'on_open'):
            self.on_open()
        self.readSettings()

    # open dialog box to choose host to connect to
    def chooseHost(self):
        userInput, ok = QInputDialog.getText(self, 'Connect',  'Server:', QLineEdit.Normal, self.connectionName)
        if ok:
            self.connectionName = userInput
            self.OpenConnection()
    
    def closeEvent(self, event):
        self.settings.setValue("geometry", self.saveGeometry())
        self.settings.setValue("windowState", self.saveState())
        if hasattr(self.__class__, 'on_close'):
            self.on_close()
        super(Gui, self).closeEvent(event)
    
    def readSettings(self):
        self.restoreGeometry(self.settings.value("geometry", QtCore.QByteArray()))
        self.restoreState(self.settings.value("windowState", QtCore.QByteArray()))

    def connectionChangedSlot(self, connected):
        self.connectionCheckbox.setChecked(connected)
        if connected:
            self.statusUpdate.emit('')
            self.connectionCheckbox.setText("Connected")
        else:
            self.connectionCheckbox.setText("NOT Connected")

    def connectionCheckedChanged(self, checked):
        self.statusUpdate.emit('')
        if checked:
            self.connectionCheckbox.setText("Connecting")
            self.OpenConnection()
        else:
            self.connectionCheckbox.setText("NOT Connected")
            self.CloseConnection()



import re
import io
import json
import os.path
import copy

class MessageException(Exception):
    pass

# create a class to load YAML files, and any YAML files they include
import yaml
class YamlLoader(yaml.Loader):
    def __init__(self, stream):
        self._dirname = os.path.dirname(stream.name)
        yaml.Loader.__init__(self, stream)

    def include(self, node):
        filename = os.path.abspath(os.path.join(self._dirname, node.value))
        try:
            inFile = io.open(filename, 'r')
            return yaml.load(inFile, YamlLoader)
        except FileNotFoundError:
            raise MessageException("Error loading " + filename + " for include statement [" + node.value + "]")

    def loadFile(self, node):
        filename = os.path.join(self._dirname, node.value)
        return "*Not* including " + filename

YamlLoader.add_constructor('!include', YamlLoader.include)
YamlLoader.add_constructor('!File', YamlLoader.loadFile)

# needs to be accessible to languages, so they can change it to true if they want
# to use non-linear conversions
use_nonlinear_conversions = False

def readFile(filename):
    #print("Processing ", filename)
    if filename.endswith(".yaml"):
        inFile = io.open(filename)
        return yaml.load(inFile, YamlLoader)
    elif filename.endswith(".json"):
        inFile = io.open(filename)
        return json.load(inFile)
    else:
        return 0

def fieldSize(field):
    fieldSizes = {"uint64":8, "uint32":4, "uint16": 2, "uint8": 1, "int64":8, "int32":4, "int16": 2, "int8": 1, "float64":8, "float32":4}
    return fieldSizes[field["Type"]]

def fieldLocation(field):
    return field['Location']

# offset between elements of an array of this field.
# same as field size for most fields, except when this is an element of an array of structs
def fieldArrayElementOffset(field):
    if 'StructSize' in field:
        return int(field['StructSize'])
    return fieldSize(field)

def msgSize(msg):
    offset = 0
    if "Fields" in msg:
        for field in msg["Fields"]:
            offset += fieldSize(field) * fieldCount(field)
    return offset

def fieldIsInt(field):
    isInt  = 0
    if "NumBits" in field:
        isInt = 1
    if "Type" in field and "int" in field["Type"]:
        isInt = 1
    return isInt

def fieldNumBits(field):
    numBits = 0
    if "Type" in field or "NumBits" in field:
        numBits = fieldItem(field, "NumBits", 0)
    if "Type" in field:
        fieldType = field["Type"]
        if "int" in fieldType:
            numBits = fieldSize(field) * 8
            if fieldType.startswith("int"):
                isSigned = 1
    return numBits

def fieldIsSigned(field):
    isSigned = 0
    if "Type" in field:
        fieldType = field["Type"]
        if "int" in fieldType:
            if fieldType.startswith("int"):
                isSigned = 1
    return isSigned

def transformInt(field, value):
    if "Scale" in field or "Offset" in field:
        scale = fieldItem(field, "Scale", 1.0)
        offset = fieldItem(field, "Offset", 0.0)
        value = float(value) * float(scale) + float(offset)
    return value

def fieldMin(field):
    if fieldIsInt(field):
        numBits = fieldNumBits(field)
        # if scale is negative, reverse min/max based on range of storage value
        if "Scale" in field and float(field['Scale']) < 0.0:
            if fieldIsSigned(field):
                minVal = 2**(numBits-1)-1
            else:
                minVal = 2**numBits-1
        else:
            if fieldIsSigned(field):
                minVal = -2**(numBits-1)
            else:
                minVal = 0
        minVal = transformInt(field, minVal)
    elif "Type" in field and field["Type"] == 'float64':
        minVal = "DBL_MIN"
    elif "Type" in field and field["Type"] == 'float32':
        minVal = "FLT_MIN"
    minVal = fieldItem(field, "Min", minVal)
    return minVal

def fieldMax(field):
    if fieldIsInt(field):
        numBits = fieldNumBits(field)
        # if scale is negative, reverse min/max based on range of storage value
        if "Scale" in field and float(field['Scale']) < 0.0:
            if fieldIsSigned(field):
                maxVal = -2**(numBits-1)
            else:
                maxVal = 0
        else:
            if fieldIsSigned(field):
                maxVal = 2**(numBits-1)-1
            else:
                maxVal = 2**numBits-1
        maxVal = transformInt(field, maxVal)
    elif "Type" in field and field["Type"] == 'float64':
        maxVal = "DBL_MAX"
    elif "Type" in field and field["Type"] == 'float32':
        maxVal = "FLT_MAX"
    maxVal = fieldItem(field, "Max", maxVal)
    return maxVal

def fieldStorageMin(storageType):
    dict = \
    {"uint64":0, "uint32":0, "uint16": 0, "uint8": 0,
      "int64": -2**63,  "int32":-2**31,  "int16": -2**15,  "int8": -2**7}
    return dict[storageType]

def fieldStorageMax(storageType):
    dict = \
    {"uint64": 2**64-1, "uint32": 2**32-1, "uint16":  2**16-1, "uint8":  2**8-1,
      "int64": 2**63-1,  "int32": 2**31-1,  "int16":  2**15-1,  "int8": 2**7-1}
    return dict[storageType]

def fieldItem(field, item, default):
    if item in field:
        return field[item]
    else:
        return default

def fieldUnits(field):
    return fieldItem(field, "Units", "")

def fieldDescription(field):
    ret = fieldItem(field, "Description", "")
    if not ret:
        ret = ""
    ret = ret.replace('\n', ' ')
    ret = ret.replace('\\', ' ')
    return ret

def fieldDefault(field):
    return fieldItem(field, "Default", "")

def fieldCount(field):
    return fieldItem(field, "Count", 1)

def numberOfFields(msg):
    count = 0
    if "Fields" in msg:
        for field in msg["Fields"]:
            count+=1
            if "Bitfields" in field:
                for bitfield in field["Bitfields"]:
                    count+=1
    return count

def numberOfSubfields(msg):
    count = 0
    if "Fields" in msg:
        for field in msg["Fields"]:
            if "Bitfields" in field:
                for bitfield in field["Bitfields"]:
                    count+=1
            else:
                count+=1
    return count

def fieldType(language, field, bitfield):
    if bitfield:
        if 'Type' in bitfield:
            type = language.fieldType(bitfield)
        else:
            type = language.fieldType(field)
        if "Offset" in field or "Scale" in bitfield:
            type = language.typeForScaledInt(field)
    else:
        type = language.fieldType(field)
        if "Offset" in field or "Scale" in field:
            type = language.typeForScaledInt(field)
    return type

def fieldReplacements(language, line, msg, subfieldsOnly):
    line = re.sub('<FOREACHFIELD\(', '', line)
    line = re.sub('<FOREACHSUBFIELD\(', '', line)
    line = re.sub('\)>$', '', line)
    ret = ""
    count = 0
    if "Fields" in msg:
        for field in msg["Fields"]:
            if not subfieldsOnly or not "Bitfields" in field:
                thisLine = line
                thisLine = thisLine.replace("<FIELDNAME>", field["Name"])
                thisLine = thisLine.replace("<FIELDNUMBER>", str(count))
                thisLine = thisLine.replace("<FIELDCOUNT>", str(fieldCount(field)))
                thisLine = thisLine.replace("<FIELDTYPE>", fieldType(language, field, None))
                ret +=  thisLine
                count+=1
            if "Bitfields" in field:
                for bitfield in field["Bitfields"]:
                    thisLine = line
                    thisLine = thisLine.replace("<FIELDNAME>", bitfield["Name"])
                    thisLine = thisLine.replace("<FIELDNUMBER>", str(count))
                    thisLine = thisLine.replace("<FIELDCOUNT>", str(fieldCount(bitfield)))
                    thisLine = thisLine.replace("<FIELDTYPE>", fieldType(language, field, None)) # bitfields inherit type of parent field
                    ret +=  thisLine
                    count+=1
    return ret 

def msgName(msg):
    try:
        return msg["Name"]
    except KeyError:
        pass
    ret = None
    # iterate over list of ids
    for id in msg["ids"]:
        subname = msg[id["Name"]]
        if "." in subname:
            subname = subname.split(".")[1]
        if ret:
            ret = ret + "_" + subname
        else:
            ret = subname
    return ret

def msgAlias(msg):
    try:
        return msg["Alias"]
    except KeyError:
        return ""
    
def msgShortName(msg):
    commonSubdir = msg["commonSubdir"]
    name = msgName(msg)
    # remove subdirs from start
    if name.startswith(commonSubdir+"_"):
        name = name.replace(commonSubdir+"_", "")
    return name

def msgID(msg, enums, undefinedMsgId):
    ret = undefinedMsgId
    # check if there was an externally specified list of ID fields
    if msg["ids"]:
        ret = 0
        # iterate over list of ids
        shiftValue = 0
        for id in msg["ids"]:
            try:
                value = msg[id["Name"]]
            except KeyError:
                value = 0
            #print("got value " + str(value))
            try:
                value = int(value)
            except ValueError:
                for enum in enums:
                    enumType = enum["Name"]
                    for option in enum["Options"]:
                        if value == enumType + "." + option["Name"] or (enum["Name"] == id["Name"]+"s" and value == option["Name"]):
                            enumName = value
                            value = int(option["Value"])
                            #print("found value " + str(value) + " for " + enumType + "." + str(enumName))
                            break
            try:
                value = int(value)
            except ValueError:
                raise MessageException("ERROR! Can't find value for " + str(value))
            shiftValue = id["Bits"]
            #print("ID " + id["Name"] + " is " + str(value) + ", " + str(id["Bits"]) + " bits")
            ret = (ret << shiftValue) + value
    if "ID" in msg:
        ret = msg["ID"]
    #print("message " + msg["Name"] + " has ID " + hex(ret))
    return str(ret)

def msgDescriptor(msg, inputFilename):
    # remove + symbols added to the path for Matlab namespaces
    inputFilename = inputFilename.replace("+","")
    subdir = msg["commonSubdir"].replace("+", "").replace("/",".")
    name = msgName(msg)
    basename = os.path.basename(inputFilename).split('.')[0]
    #print("[%s] [%s] [%s] [%s]" % (inputFilename, subdir, name, basename))
    # add the filename as a namespace, unless it's already there
    if not basename == name and not basename+'_' in name:
        name = basename + '.' + name
        #print("adding basename " + basename + " to " + name)
    if name.startswith(subdir+"_"):
        return name.replace("_", ".")
    if subdir and not name.startswith(subdir):
        return subdir+"."+name.replace("_", ".")
    return name

def addShift(base, value, shiftValue):
    ret = value
    if base != "":
        ret = "(("+base+")<<"+str(shiftValue)+")"+"+"+value
    return ret
    
def baseGetMsgID(prefix, baseParam, typecast, enumAsIntParam, msg):
    ret = ""
    if "Fields" in msg:
        for field in msg["Fields"]:
            if "IDBits" in field:
                numBits = field["IDBits"]
                param = baseParam
                if "Enum" in field and enumAsIntParam:
                    if param != "":
                        param += ", "
                    param += "1"
                getStr = prefix+"Get"+field["Name"]+"("+param+")"
                if typecast == "CAST_ALL" or ("Enum" in field and typecast != 0):
                    getStr = "(uint32_t)("+getStr+")"
                ret =  addShift(ret, getStr, numBits)
            if "Bitfields" in field:
                for bitfield in field["Bitfields"]:
                    if "IDBits" in bitfield:
                        numBits = bitfield["IDBits"]
                        param = baseParam
                        if "Enum" in bitfield and enumAsIntParam:
                            if param != "":
                                param += ", "
                            param += "1"
                        getStr = prefix+"Get"+BitfieldName(field, bitfield)+"("+param+")"
                        if typecast == "CAST_ALL" or ("Enum" in bitfield and typecast != 0):
                            getStr = "(uint32_t)("+getStr+")"
                        ret =  addShift(ret, getStr, numBits)
    return ret
    
def baseSetMsgID(prefix, param, castEnums, enumAsIntParam, msg):
    ret = ""
    numBits = 0
    if "Fields" in msg:
        for field in reversed(msg["Fields"]):
            if "IDBits" in field:
                if numBits != 0:
                    ret += "\nid = id >> " + str(numBits)+"\n"
                numBits = field["IDBits"]
                setStr = "id & "+Mask(numBits)
                if "Enum" in field and castEnums:
                    setStr = field["Enum"]+"("+setStr+")"
                ret +=  prefix+"Set"+field["Name"]+"("+param+setStr+")"
            if "Bitfields" in field:
                for bitfield in reversed(field["Bitfields"]):
                    if "IDBits" in bitfield:
                        if numBits != 0:
                            ret += "\nid = id >> " + str(numBits)+"\n"
                        numBits = bitfield["IDBits"]
                        setStr = "id & "+Mask(numBits)
                        if "Enum" in bitfield and castEnums:
                            setStr = bitfield["Enum"]+"("+setStr+")"
                        ret +=  prefix+"Set"+bitfield["Name"]+"("+param+setStr+")"
    return ret

# return a list of all enumerations in this input file, or anything it includes
def Enums(inputData):
    enumList = []
    if "Enums" in inputData:
        enumList = inputData["Enums"]
    if "includes" in inputData:
        for data in inputData["includes"]:
            enumList = enumList + Enums(data)
    return enumList

def Structs(inputData):
    structList = {}
    if "Structs" in inputData:
        structList = inputData["Structs"]
    if "includes" in inputData:
        for data in inputData["includes"]:
            structList.update(Structs(data))
    return structList

def NestedArraysExist(field, structs, outer_array_exists):
    if fieldCount(field) > 1:
        if outer_array_exists:
            return True
        outer_array_exists = True
    if field["Type"] in structs:
        s = structs[field["Type"]]
        for subfield in s['Fields']:
            found_sub_array = NestedArraysExist(subfield, structs, outer_array_exists)
            if found_sub_array:
                return True
    return False
    
# this replaces fields that are references to structs with the fields from the referenced struct,
# and also sets a few items of field meta-data (Location, StructSize, Count)
def PatchStructs(inputData):
    # loop twice, so that references to structs inside structs are also replaced
    for i in range(2):
        structs = Structs(inputData)
        #if not structs:
        #    return
        
        # need to make a new list, because we'll be inserting elements as we iterate
        if 'Messages' in inputData:
            for msg in inputData["Messages"]:
                resolve_sizes_and_locations = True
                if 'Fields' in msg:
                    location = 0
                    outfields = []
                    for field in msg['Fields']:
                        nested_structs = NestedArraysExist(field, structs, False)
                        if nested_structs:
                            #print("found nested struct in %s.%s" % (msg["Name"], field["Name"]))
                            if fieldCount(field) > 1:
                                # flatten array!
                                #print("Flattening array %s[%d]" % (field["Name"], fieldCount(field)))
                                for j in range(fieldCount(field)):
                                    fieldcopy = copy.deepcopy(field)
                                    fieldcopy['Name'] = fieldcopy['Name'] + "_" + str(j)
                                    fieldcopy['Count'] = 1
                                    outfields.append(fieldcopy)
                                continue
                        if field["Type"] in structs:
                            s = structs[field["Type"]]
                            # For arrays of structs, compute the size of the structure and store it in each
                            # field in the structure.  This is useful for finding subsequent elements of
                            # fields of arrays of structs, because we'll treat them as parallel arrays in
                            # the generated code, not as hierarchical structs.
                            structSize = 0
                            for subfield in s['Fields']:
                                if resolve_sizes_and_locations:
                                    try:
                                        fs = fieldSize(subfield) * fieldCount(subfield)
                                        if fieldCount(field) > 1 or fieldCount(s) > 1:
                                            structSize += fs
                                    except KeyError:
                                        # if we have a KeyError when trying to get fieldSize,
                                        # it's because we have a nested Struct that hasn't been
                                        # replaced by it's members yet.  Don't worry about that now,
                                        # because we'll iterate and replace it in the next loop.
                                        resolve_sizes_and_locations = False
                                        structSize = None
                                        location = None
                            # track field offsets within struct
                            subfieldLocation = 0
                            for subfield in s['Fields']:
                                subfieldcopy = copy.deepcopy(subfield)
                                #  if a field name was specified, prepend it to the names of the sub-fields
                                # of the struct.  Also allow an unnamed field so that nothing is prepended.
                                if 'Name' in field:
                                    subfieldcopy['Name'] = field['Name'] + "_" + subfield['Name']
                                else:
                                    subfieldcopy['Name'] = subfield['Name']
                                if resolve_sizes_and_locations:
                                    if structSize != 0:
                                        subfieldcopy['StructSize'] = structSize
                                    subfieldcopy['Location'] = location + subfieldLocation
                                # subfield count should be greater of parent field count and subfield type count
                                # error if both are > 1
                                if fieldCount(subfieldcopy) == 1:
                                    if fieldCount(field) > 1:
                                        subfieldcopy['Count'] = fieldCount(field)
                                else:
                                    if fieldCount(field) > 1:
                                        raise MessageException("ERROR! %s Field %s Count %d > 1 and Subfield %s Count %d > 1" % (msg["Name"], field['Name'], fieldCount(field), subfieldcopy['Name'], fieldCount(subfieldcopy)))
                                        subfieldcopy['Count'] = fieldCount(field) * fieldCount(subfieldcopy)
                                #TODO How to handle array of bitfields?!?
                                if "Bitfields" in subfieldcopy:
                                    for bits in subfieldcopy["Bitfields"]:
                                        bits['Name'] = field['Name'] + "_" + bits['Name']
                                outfields.append(subfieldcopy)
                                if resolve_sizes_and_locations:
                                    subfieldLocation += fieldSize(subfield) * fieldCount(subfield)
                            if resolve_sizes_and_locations:
                                # add size of struct times struct count to running total of field location
                                location += subfieldLocation * fieldCount(field)
                        else:
                            outfields.append(field)
                            if resolve_sizes_and_locations:
                                if not 'Location' in field:
                                    field['Location'] = location
                                location += fieldSize(field) * fieldCount(field)
                    msg['Fields'] = outfields

# sanitize the option name, to have valid identifier characters
def OptionName(option):
    return option["Name"].replace("/", "_or_").replace(" ", "_").replace("-", "_")

# return a list of all IDs in this input file, or anything it includes
def MsgIDs(inputData):
    idList = []
    if "IDs" in inputData:
        idList = inputData["IDs"]
    if "includes" in inputData:
        for data in inputData["includes"]:
            idList = idList + MsgIDs(data)
    return idList

# return a list of just the enums that are used by a message's fields/bitfields
# this is useful for languages that put enum info into their output file, because
# it will be a list of only the enums that are relevant.  often there can be many
# enums defiined in a common include file, and they aren't all used by a particular
# message
def UsedEnums(msg, enums):
    usedEnums = []
    for enum in enums:
        if "Fields" in msg:
            for field in msg["Fields"]:
                if "Enum" in field:
                    if field["Enum"] == enum["Name"]:
                        usedEnums.append(enum)
                        break
                foundEnum = 0
                if "Bitfields" in field:
                    for bits in field["Bitfields"]:
                        if "Enum" in bits:
                            if bits["Enum"] == enum["Name"]:
                                usedEnums.append(enum)
                                foundEnum = 1
                                break
                    if foundEnum:
                        break

    try:
        if msg["ids"]:
            idEnum = {}
            idEnum["Name"] = "IDs"
            idEnum["Options"] = []
            for id in msg["ids"]:
                idName = id["Name"]
                idValue = msg[idName]
                try:
                    idValue = int(idValue)
                except ValueError:
                    for enum in enums:
                        enumType = enum["Name"]
                        for option in enum["Options"]:
                            if idValue == enumType + "." + option["Name"] or (enum["Name"] == id["Name"]+"s" and idValue == option["Name"]):
                                enumName = idValue
                                idValue = int(option["Value"])
                                break
                try:
                    idValue = int(idValue)
                except ValueError:
                    raise MessageException("ERROR! Can't find value for " + str(idValue))
                option = {}
                option["Name"] = idName
                option["Value"] = idValue
                idEnum["Options"].append(option)
            usedEnums.append(idEnum)
    except KeyError:
        pass
    return usedEnums

def typeForScaledInt(field):
    numBits = fieldNumBits(field)
    if numBits > 24:
        return "double"
    return "float"

# for floats, append tag such as 'f' to constants to eliminate compiler warnings
def fieldScale(field, floatTag):
    if "Scale" in field:
        ret = field["Scale"]
        if typeForScaledInt(field) == "float":
            ret = str(ret) + floatTag
    return ret

def fieldOffset(field, floatTag):
    if "Offset" in field:
        ret = field["Offset"]
        if typeForScaledInt(field) == "float":
            ret = str(ret) + floatTag
    return ret

def conversionParams(conversion, conversionParamNames):
    c = ""
    for param in conversion:
        if param != "Type":
            if conversionParamNames:
                c += "%s=" % (param)
            c += "%s, " % (conversion[param])
    return c[:-2]

# add tag (such as 'f') to scale and offset for floats
def getMath(x, field, cast, floatTag="", conversionParamNames=False):
    ret = x
    if cast and ("Offset" in field or "Scale" in field):
        ret = "%s(%s)" % (cast, ret)
    if "Scale" in field:
        ret = "(%s * %s)" % (ret, fieldScale(field, floatTag))
    if "Offset" in field:
        ret = "(%s + %s)" % (ret, fieldOffset(field, floatTag))
    if use_nonlinear_conversions and "Conversion" in field:
        conversion = field["Conversion"]
        ret = "%s.Convert(%s, %s)" % (conversion["Type"], ret, conversionParams(conversion, conversionParamNames))
    return ret

def setMath(x, field, cast, floatTag="", conversionParamNames=False):
    ret = x
    if "Offset" in field:
        ret = "(%s - %s)" % (ret, fieldOffset(field, floatTag))
    if "Scale" in field:
        ret = "%s / %s" % (ret, fieldScale(field, floatTag))
    if cast and ("Offset" in field or "Scale" in field):
        ret = "%s(%s)" % (cast, ret)
    if use_nonlinear_conversions and "Conversion" in field:
        conversion = field["Conversion"]
        ret = "%s.Invert(%s, %s)" % (conversion["Type"], ret, conversionParams(conversion, conversionParamNames))
    return ret

def fieldHasConversion(field):
    return "Offset" in field or "Scale" in field or ("Conversion" in field and use_nonlinear_conversions)

def Mask(numBits):
    return str(hex(2 ** numBits - 1))

def BitfieldName(field, bits):
    #return str(field["Name"]) +str(bits["Name"])
    return str(bits["Name"])
        



import msgtools.parser.parser as MsgParser

def accessors(msg):
    return ""

def createFieldInfoRow(field):
    name = str(field["Name"]) if "Name" in field else "ERROR"
    fieldtype = str(field["Type"]) if "Type" in field else "ERROR"
    count = str(field["Count"]) if "Count" in field else "1"

    units = str(field["Units"]) if "Units" in field else ""

    if "Enum" in field:
        units = str(field["Enum"])
    elif "Bitfields" in field:
        units = "Bitfield"

    scale = str(field["Scale"]) if "Scale" in field else ""
    offset = str(field["Offset"]) if "Offset" in field else ""
    default = str(field["Default"]) if "Default" in field else ""
    min = str(MsgParser.fieldMin(field))
    max = str(MsgParser.fieldMax(field))
    description = str(field["Description"]) if "Description" in field else ""

    html = "\
        <tr>\
            <td class='fieldname'>" + name + "</td>\
            <td class='fieldtype'>" + fieldtype + "</td>\
            <td class='fieldcount'>" + count + "</td>\
            <td class='fieldunits'>" + units + "</td>\
            <td class='fieldscale'>" + scale + "</td>\
            <td class='fieldoffset'>" + offset + "</td>\
            <td class='fielddefault'>" + default + "</td>\
            <td class='fielddefault'>" + min + " - " + max + "</td>\
            <td class='fielddescription'>" + description + "</td>\
        </tr>\
    "
    return html

def createBitfieldInfoRow(bitfield):
    name = "&nbsp&nbsp&nbsp&nbsp" + str(bitfield["Name"]) if "Name" in bitfield else "ERROR"
    bitfieldtype = str(bitfield["NumBits"]) + " bits" if "NumBits" in bitfield else "ERROR"
    count = str(bitfield["Count"]) if "Count" in bitfield else "1"
    units = str(bitfield["Units"]) if "Units" in bitfield else ""

    if "Enum" in bitfield:
        units = "<ENUM>" + str(bitfield["Enum"]) + "</ENUM>"

    scale = str(bitfield["Scale"]) if "Scale" in bitfield else ""
    offset = str(bitfield["Offset"]) if "Offset" in bitfield else ""
    default = str(bitfield["Default"]) if "Default" in bitfield else ""
    min = str(MsgParser.fieldMin(bitfield))
    max = str(MsgParser.fieldMax(bitfield))
    description = str(bitfield["Description"]) if "Description" in bitfield else ""

    html = "\
        <tr>\
            <td> <span style=\"font-size: 1.3em;\">&#x21e8;</span>" + name + " </td>\
            <td>" + bitfieldtype + "</td>\
            <td>" + count + "</td>\
            <td>" + units + "</td>\
            <td>" + scale + "</td>\
            <td>" + offset + "</td>\
            <td>" + default + "</td>\
            <td>" + min + " - " + max + "</td>\
            <td>" + description + "</td>\
        </tr>\
    "
    return html

def initCode(msg):
    ret = []

    if "Fields" in msg:
        for field in msg["Fields"]:
            fieldInfoRow = createFieldInfoRow(field)
            
            if fieldInfoRow:
                ret.append(fieldInfoRow)
            
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    bits_html = createBitfieldInfoRow(bits)
                    if bits:    
                        ret.append(bits_html)

    return ret

def enums(e):
    html = []

    for enum in e:
        enum["Name"]

        options = []

        for option in enum["Options"]:
            options.append("<tr><td>" + str(option["Name"]) + "</td><td>" + str(option["Value"]) + "</td></tr>")

        html.append("<div class='row'>\
            <div class='col-lg-4'>\
                <table class='table table-bordered table-hover table-striped'>\
                    <caption>Enumeration: " + enum["Name"] + "</caption>\
                    <thead>\
                        <tr>\
                            <th>Option</th>\
                            <th>Value</th>\
                        </tr>\
                    </thead>\
                    " + "\n".join(options) + "\
                </table>\
            </div>\
        </div>")

    return "\n".join(html)

def declarations(msg):
    return [""]



#!/usr/bin/env python3
import os
import sys
import argparse
import json
import shutil
import jinja2

apppath = os.path.split(os.path.realpath(__file__))[0]

description = '''Build a base MsgTools web app.  This utility will inspect a given generated code
    directory and generate a boilerplate web app that initializes msgtools, and loads all the
    messages defined in the given message directory.  The utility also copies over the bundled
    version of msgtools.js, plotting, and ui components into your new app directory.  This
    keeps the parser Javascript templates in sync with MsgTools libraries so changes to templates
    or the libraries don't break across versions.

    Jinja2 is used to generate the final web app.  Command line arguments are
    provided that allow you to replace the default template provided by MsgTools
    with your own and provide template arguments in the form of key/value pairs.
    Required command line arguments cover template arguments we deem essential.
    For your own templates we require:
        {{webdir}}   - a web friendly base path to generated message code
        {{messages}} - an array of messages files to load.  Autogenerated from the generated
                       code directory.
        {{appname}} - The name of the app'''


def discoverMessages(msgdir):
    '''Iterate the message directory and subdirectory to identify all .js
    files and build a list of all message "modules" within the directory

    return that list'''

    # Make sure msgdir ends with a slash to simplify things later
    if msgdir[len(msgdir) - 1] != os.sep:
        msgdir += os.sep

    messages = []
    dirs = [msgdir]
    while len(dirs) is not 0:
        currentDir = dirs.pop()
        for entry in os.listdir(currentDir):
            currentPath = os.path.join(currentDir, entry)
            if os.path.isdir(currentPath):
                dirs.append(currentPath)
            elif entry.endswith('.js'):
                # Drop the base msgdir and extension
                message = currentPath[
                    0:currentPath.rfind('.')][len(msgdir):]
                message = message.replace(os.sep, '.')
                messages.append(message)
            else:
                print('Skipping {0}; not a Javascript file'.format(
                    currentPath))

    return messages


def generateAppFile(templatepath, jinjaArgs, outputfile):
    '''Apply the template args to the template and dump the result into
    the output file
    templatepath should be a filepath
    jinjaArgs should be a string to value dictionary
    outputfile should be an output filepath
    '''
    with open(templatepath, 'r') as fp:
        content = fp.read()

    template = jinja2.Template(content)
    rendering = template.render(**jinjaArgs)

    with open(outputfile, 'w') as fp:
        fp.write(rendering)


def copyFiles(srcpath, destpath):
    '''Shallow copy of the src path to the destination path'''
    for file in os.listdir(srcpath):
        src = os.path.join(srcpath, file)
        dest = os.path.join(destpath, file)
        shutil.copy2(src, dest)


def buildApp(htmlTemplate, jsTemplate, jinjaArgs, msgdir, outputdir):
    '''We're  just going to iterate each file in the directory and build a
    message entry for it. Then we'll run Jinja to build a custom web app based
    on the messages we've processed'''

    print('Analyzing message directory...')
    jinjaArgs['messages'] = discoverMessages(msgdir)
    jinjaArgs['appjs'] = jinjaArgs['appname'].replace(' ', '_').lower() + '.js'

    print('Generating {0}...'.format(os.path.join(outputdir, 'index.html')))
    generateAppFile(htmlTemplate, jinjaArgs,
                    os.path.join(outputdir, 'index.html'))

    # Generate the app js file
    jsOutputname = os.path.join(outputdir, jinjaArgs['appjs'])
    print('Generating {0}...'.format(jsOutputname))
    generateAppFile(jsTemplate, jinjaArgs, jsOutputname)

    # Copy msgtools libraries to the output directory
    # Note that you should also add these files to setup.py: package data
    # so they are included with the pip install package.  Otherwise this
    # call will fail...
    print('Copying library files...')
    copyFiles(os.path.join(apppath, 'lib'), outputdir)


def getTemplate(templatePath, defaultTemplate):
    retVal = None
    if templatePath is None:
        retVal = os.path.join(apppath, defaultTemplate)
    elif os.path.exists(templatePath) and os.path.isfile(templatePath):
        retVal = templatePath
    else:
        print('{0} does not exist, or is a directory'.format(templatePath))
        sys.exit(1)

    return retVal


def main():
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('-w', '--web', dest='webdir', required=False,
                        help='The web friendly msg path inserted into the app template.  Defaults to msgdir if not specified.')
    parser.add_argument('-m', '--html', dest='htmlTemplate',
                        help='Use the specified HTML template instead of the default for building the web app')
    parser.add_argument('-j', '--javascript', dest='jsTemplate',
                        help='Use the specified Javascript template instead of the default')
    parser.add_argument('-t', '--templateargs', dest='templateArgs', type=str,
                        help='''Arguments to pass to the template engine as a set of key/value pairs. This argument is first interpreted
                as a path to a JSON formatted file.  If it is not a valid file, then the argument is treated as a
                JSON string.  Literal JSON example: \'{"key":"value"}\' ''')
    parser.add_argument('appname', help='The name of the app.')
    parser.add_argument(
        'msgdir', help='The basepath for where generated message code is placed.  For example, obj/CodeGenerator/Javascript')
    parser.add_argument(
        'outputdir', help='The destination directory for the resulting HTML app.  Defaults to the current directory.')

    parser.add_argument("-d", "--disable-ui",
                        help='''If specified, disables generation of UI components for socket connection and logging,
                        in favor of a messaging client object. [default enabled]''',
                        dest='widgets',
                        action='store_false',
                        default=True
                        )

    args = parser.parse_args()

    print('Preparing environment...')

    # Setup our templates
    htmlTemplate = getTemplate(args.htmlTemplate, 'template.html')
    jsTemplate = getTemplate(args.jsTemplate, 'template.js')

    # Setup our template arguments
    jinjaArgs = {}
    if args.templateArgs is not None:
        if os.path.exists(args.templateArgs) is True:
            if os.path.isdir(args.templateArgs):
                print('{0} is a directory.'.format(args.templateArgs))
                sys.exit(1)
            elif os.path.exists(args.templateArgs) is True:
                with open(args.templateArgs) as fp:
                    try:
                        jinjaArgs = json.load(fp)
                    except Exception as e:
                        print('{0} not valid json'.format(args.templateArgs))
                        print(e)
                        sys.exit(1)
        else:
            try:
                jinjaArgs = json.loads(args.templateArgs)
            except Exception as e:
                print('{0} not valid json'.format(args.templateArgs))
                print(e)
                sys.exit(1)

    jinjaArgs['appname'] = args.appname
    jinjaArgs['webdir'] = args.msgdir if args.webdir is None else args.webdir
    jinjaArgs['widgets'] = args.widgets

    # Verify the message basepath exists
    if os.path.exists(args.msgdir) is False or os.path.isdir(args.msgdir) is False:
        print('{0} does not exist, or is not a directory'.format(args.msgdir))
        sys.exit(1)

    # Setup our destination dir
    if os.path.exists(args.outputdir) is False or os.path.isdir(args.outputdir) is False:
        os.mkdir(args.outputdir)

    buildApp(htmlTemplate, jsTemplate, jinjaArgs, args.msgdir, args.outputdir)

    print('DONE')

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import os
import sys
import argparse
from PyQt5 import QtCore, QtGui, QtWidgets

# if started via invoking this file directly (like would happen with source sitting on disk),
# insert our relative msgtools root dir into the sys.path, so *our* msgtools is used, not
# any other already in the path.
if __name__ == '__main__':
    srcroot=os.path.abspath(os.path.dirname(os.path.abspath(__file__))+"/../..")
    sys.path.insert(1, srcroot)
from msgtools.lib.messaging import Messaging
import msgtools.lib.gui

DESCRIPTION='''MsgInspector allows you to connect to a MsgServer and inspect messages as they arrive.
    It is similar to MsgScope but presents data in a time linear format with compact details.  Each message
    type grouped into it's own tab'''

class MsgInspector(msgtools.lib.gui.Gui):
    def __init__(self, parent=None):
        parser = argparse.ArgumentParser(description=DESCRIPTION)
        parser.add_argument('--sortfields', nargs='+', help='''One or more msgName/fieldName specifiers used to indicate
            which field on a message should be used for sorting. Example --sortfields Printf/LineNumber.  If --msgs is specified,
            and a message name is requested that isn\'t in the whitelist, the message will be added to the whitelist for you.''')
        parser = msgtools.lib.gui.Gui.addBaseArguments(parser)
        args=parser.parse_args()

        msgtools.lib.gui.Gui.__init__(self, "Message Inspector 0.1", args, parent)

        self.keyFields = {}

        # Process our msg/keyfields
        if args.sortfields is not None:
            for sortfield in args.sortfields:
                try:
                    msgName,fieldName = sortfield.split('/')
                except Exception:
                    print('--sortfields {0} is invalid.  Specify a message name and field name separated by a forward slash.'.format(sortfield))
                    sys.exit(1)

                if msgName not in Messaging.MsgIDFromName:
                    print('{0} is an unknown message name.'.format(msgName))
                    sys.exit(1)

                if isinstance(fieldName, str) is False:
                    print(msgfield, ': You may only specify one field name per message.')
                    sys.exit(1)

                MsgClass = Messaging.MsgClassFromName[msgName]
                if hasattr(MsgClass, 'fields'):
                    fieldNames = [field.name for field in MsgClass.fields]
                    if fieldName not in fieldNames:
                        print('{0} is not a field of {1}.  Valid fields are: {2}'.format(fieldName, msgName,
                            fieldNames))
                        sys.exit(1)

                # If we have a message white list then add this message to the set
                # Length of zero means all messages are being accepted.
                if len(self.allowedMessages) > 0:
                    self.allowedMessages.add(msgName)
                self.keyFields[msgName] = fieldName

        # event-based way of getting messages
        self.RxMsg.connect(self.ProcessMessage)

        # tab widget to show multiple messages, one per tab
        self.tabWidget = QtWidgets.QTabWidget(self)
        self.setCentralWidget(self.tabWidget)
        
        # hash table to lookup the widget for a message, by message ID
        self.msgWidgets = {}
        
        # whether we should autoscroll as data is added
        self.autoscroll = 1
        
        # menu items to change view
        clearAction = QtWidgets.QAction('&Clear', self)
        clearAllAction = QtWidgets.QAction('Clear &All', self)
        self.scrollAction = QtWidgets.QAction('&Scroll', self)
        self.scrollAction.setCheckable(1)
        self.scrollAction.setChecked(self.autoscroll)

        menubar = self.menuBar()
        viewMenu = menubar.addMenu('&View')
        viewMenu.addAction(clearAction)
        viewMenu.addAction(clearAllAction)
        viewMenu.addAction(self.scrollAction)

        clearAction.triggered.connect(self.clearTab)
        clearAllAction.triggered.connect(self.clearAllTabs)
        self.scrollAction.triggered.connect(self.switchScroll)
    
    def clearTab(self):
        self.tabWidget.currentWidget().clear()
    
    def clearAllTabs(self):
        for id, widget in self.msgWidgets.items():
            widget.clear()
        
    def switchScroll(self):
        self.autoscroll = not self.autoscroll
        self.scrollAction.setChecked(self.autoscroll)

    def tableDataDoubleClicked(self, treeWidgetItem, column):
        self.autoscroll = not self.autoscroll
        self.scrollAction.setChecked(self.autoscroll)

    def ProcessMessage(self, msg):
        hdr = msg.hdr
        
        # create identifying string for type and source of message
        msgRoute = Messaging.MsgRoute(msg)
        msgKey = msg.MsgName()
        if len(msgRoute) > 0 and not(all ("0" == a for a in msgRoute)):
            msgKey += " ("+",".join(msgRoute)+")"

        if(not(msgKey in self.msgWidgets)):
            # create a new tree widget
            keyField = self.keyFields.get(msg.MsgName(), None)
            msgWidget = msgtools.lib.gui.MsgTreeWidget(type(msg), keyField)

            # connect to double-clicked signal to change scrolling globally
            msgWidget.itemDoubleClicked.connect(self.tableDataDoubleClicked)
            
            # add it to the tab widget, so the user can see it
            self.tabWidget.addTab(msgWidget, msgKey)
            
            # store a pointer to it, so we can find it next time (instead of creating it again)
            self.msgWidgets[msgKey] = msgWidget
        
        # give the data to the widget
        self.msgWidgets[msgKey].addData(msg, self.autoscroll)

def main(args=None):
    app = QtWidgets.QApplication(sys.argv)
    msgApp = MsgInspector()
    msgApp.show()
    sys.exit(app.exec_())

# main starts here
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
Plot message data in scrolling window
"""
import os
import sys
import argparse

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *

import pyqtgraph as pg
from pyqtgraph.Qt import QtCore, QtGui
import numpy as np

# if started via invoking this file directly (like would happen with source sitting on disk),
# insert our relative msgtools root dir into the sys.path, so *our* msgtools is used, not
# any other already in the path.
if __name__ == '__main__':
    srcroot=os.path.abspath(os.path.dirname(os.path.abspath(__file__))+"/../..")
    sys.path.insert(1, srcroot)
from msgtools.lib.messaging import Messaging

from datetime import datetime
from datetime import timedelta

from collections import namedtuple

# make tuple for line, and have multiple of them in a MsgPlot, as long as their units match and they are from the same message
LineInfo = namedtuple('LineInfo', 'msgClass msgKey baseName fieldInfo fieldSubindex dataArray timeArray curve ptr1')

start_time = datetime.now().timestamp()
def elapsedSeconds(timestamp):
    if timestamp > start_time:
        return timestamp - start_time
    return timestamp

# I just added an optimization that replaces sequences of flat data
# with just two points, so we're starting to be a little smarter
# about storing data.  maybe we want to decimate very old data?
# that'll look wrong if user zooms in, though
# it also breaks time slider, because there's no longer the
# same number of points in each line for the same time period :(
# timeslider should likely change to using actual X axis ranges,
# and then we need to find points that falls in that time in the data?
# if time resets (which happens often when msgserver restarts or
# embedded processor resets depending on how time is managed,
# then time is non-linear and not necessarily always increasing in
# the arrays and any kind of smart search won't work well :(
class MsgPlot(QWidget):
    class PlotError(Exception):
        pass

    class NewPlotError(PlotError):
        pass

    Paused = QtCore.pyqtSignal(bool)
    AddLineError = QtCore.pyqtSignal(str)
    RegisterForMessage = QtCore.pyqtSignal(str)
    MAX_LENGTH = 1000
    def __init__(self, msgClass, msgKey, fieldName, runButton = None, clearButton = None, timeSlider = None, displayControls=True, fieldLabel=None):
        super(QWidget,self).__init__()
        
        newFieldName, fieldIndex = MsgPlot.split_fieldname(fieldName)
        fieldInfo = Messaging.findFieldInfo(msgClass.fields, newFieldName)
        if fieldInfo == None:
            raise MsgPlot.PlotError("Invalid field %s for message %s" % (newFieldName, msgClass.MsgName()))

        layout = QVBoxLayout()
        self.setLayout(layout)
        #self.msgClass = msgClass
        self.pause = 0
        self.lineCount = 0
        self.showUnitsOnLegend = True
        self.units = fieldInfo.units
        self.lines = []

        yAxisLabel = fieldInfo.units
        xAxisLabel = "time (s)"
        self.plotWidget = pg.PlotWidget(labels={'left':yAxisLabel,'bottom':xAxisLabel})
        layout.addWidget(self.plotWidget)
        self.plotWidget.addLegend()
        self.addLine(msgClass, msgKey, fieldName, fieldLabel)

        # set up click handler to pause graph
        self.plotWidget.scene().sigMouseClicked.connect(self.mouseClicked)
        
        hLayout = QHBoxLayout()
        layout.addLayout(hLayout)
        
        # add a Pause/Run button
        if runButton == None:
            self.runButton = QPushButton("Pause")
        else:
            self.runButton = runButton
        self.runButton.clicked.connect(self.pauseOrRun)
        if displayControls:
            hLayout.addWidget(self.runButton)

        # add a 'Clear' button
        if clearButton == None:
            self.clearButton = QPushButton("Clear")
        else:
            self.clearButton = clearButton
        self.clearButton.clicked.connect(self.clearData)
        if displayControls:
            hLayout.addWidget(self.clearButton)

        # add slider bar to control time scale
        if timeSlider == None:
            self.timeSlider = QSlider(Qt.Horizontal)
            self.timeSlider.setMinimum(50)
            self.timeSlider.setMaximum(MsgPlot.MAX_LENGTH)
            self.timeSlider.setSingleStep(10)
            self.timeSlider.setPageStep(50)
        else:
            self.timeSlider = timeSlider
        self.timeSlider.valueChanged.connect(self.timeScaleChanged)
        if displayControls:
            hLayout.addWidget(QLabel("Time Scale"))
            hLayout.addWidget(self.timeSlider)
        
        self.plotWidget.dragEnterEvent = self.dragEnterEvent
        self.plotWidget.dragMoveEvent = self.dragMoveEvent
        self.plotWidget.dropEvent = self.dropEvent
        self.plotWidget.setAcceptDrops(1)

    @staticmethod
    def split_fieldname(fieldName):
        fieldIndex = None
        if '[' in fieldName  and ']' in fieldName:
            splits = fieldName.split('[')
            splits[1] = splits[1].replace(']','')
            fieldName = splits[0]
            fieldIndex = int(splits[1])
        return (fieldName, fieldIndex)

    def dragEnterEvent(self, ev):
        # need to accept enter event, or we won't get move event
        ev.accept()

    def dragMoveEvent(self, ev):
        # need to accept move event, or we won't get drop event
        ev.accept()

    def dropEvent(self, ev):
        ev.accept()
        item = ev.source().currentItem()
        try:
            # add a line for whatever got dropped on us
            msgClass = type(item.msg)
            self.addLine(msgClass, item.msg_key, item.fieldName)
            # register to receive message updates for the new line
            self.RegisterForMessage.emit(item.msg_key)
        except MsgPlot.PlotError as e:
            self.AddLineError.emit(str(e))
    
    def clearData(self):
        for line in self.lines:
            line.dataArray.clear()
            line.timeArray.clear()
            self.refreshLine(line)

    def addLine(self, msgClass, msgKey, fieldName, fieldLabel = None):
        fieldName, fieldIndex = MsgPlot.split_fieldname(fieldName)
        fieldInfo = Messaging.findFieldInfo(msgClass.fields, fieldName)
        if fieldInfo == None:
            raise MsgPlot.PlotError("Invalid field %s for message %s" % (fieldName, msgClass.MsgName()))
        
        if fieldInfo.units == "ASCII":
            raise MsgPlot.PlotError("Cannot plot %s.%s, it is a string" % (msgClass.MsgName(), fieldName))
        
        # don't add if it's already there!
        for line in self.lines:
            if fieldInfo == line.fieldInfo and fieldIndex == line.fieldSubindex:
                name = fieldInfo.name
                if fieldInfo.count > 1:
                    name = "%s[%d]" % (name, fieldIndex)
                raise MsgPlot.PlotError("Line %s already on plot" % name)
        #TODO why not plot things from different messages?!?
        #if msgClass != self.msgClass:
        #    raise MsgPlot.NewPlotError("Message %s != %s, cannot add to same plot" % (msgClass.__name__, self.msgClass.__name__))
        if self.showUnitsOnLegend:
            if fieldInfo.units != self.units:
                self.showUnitsOnLegend = True
                self.units = None
                #TODO how to edit existing left axis label?!?
                #print(str(self.plotWidget)) # how to print all attributes of an object?
                # pyqtgraph.widgets.PlotWidget.PlotWidget just shows up as <pyqtgraph.widgets.PlotWidget.PlotWidget object at 0x7f64c4644168>
                #self.plotWidget.labels['left'] = 'none?!?'
                for line in self.lines:
                    pass
                    #TODO how to edit existing legend
                    #line.curve.setName(line.baseName)

        if fieldIndex != None:
            self._addLine(msgClass, msgKey, fieldInfo, fieldIndex, fieldLabel, self.showUnitsOnLegend)
        elif fieldInfo.count == 1:
            self._addLine(msgClass, msgKey, fieldInfo, 0, fieldLabel, self.showUnitsOnLegend)
        else:
            dups = []
            for fieldIndex in range(0, fieldInfo.count):
                duplicate = False
                for line in self.lines:
                    if fieldInfo == line.fieldInfo and fieldIndex == line.fieldSubindex:
                        dups.append(fieldIndex)
                        duplicate = True
                        break
                if not duplicate:
                    self._addLine(msgClass, msgKey, fieldInfo, fieldIndex, fieldLabel, self.showUnitsOnLegend)
            if len(dups) > 0:
                if len(dups) == 1:
                    s = ' '+str(dups[0])
                elif len(dups) == fieldInfo.count:
                    s = 's %d-%d' % (0, fieldInfo.count)
                else:
                    s = 's '
                    for d in dups:
                        s += '%s,' % d
                raise MsgPlot.PlotError("Line%s already on plot" % s)

    def _addLine(self, msgClass, msgKey, fieldInfo, fieldIndex, fieldLabel = None, showUnits=False):
        lineName = fieldInfo.name
        baseName = lineName
        if showUnits and fieldInfo.units != '' and fieldInfo.units != 'UNKNOWN':
            lineName = lineName + " (%s)" % fieldInfo.units
        if fieldLabel != None:
            lineName = fieldLabel
            baseName = lineName
        try:
            if fieldInfo.count != 1:
                lineName += "["+str(fieldIndex)+"]"
        except:
            pass
        dataArray = []
        timeArray = []
        ptr1 = 0
        self.useHeaderTime = 0
        # This is ugly, but try to make plotWidget.plot(pen) parameters that result in colors easy to distinguish.
        line_number = len(self.lines)
        line_count_estimate = 6
        while line_count_estimate < len(self.lines)+1:
            line_count_estimate = line_count_estimate + 6
            line_number = (line_number - 6)*2+1
        curve = self.plotWidget.plot(timeArray, dataArray, name=lineName, pen=(line_number,line_count_estimate))
        lineInfo = LineInfo(msgClass, msgKey, baseName, fieldInfo, fieldIndex, dataArray, timeArray, curve, ptr1)
        self.lines.append(lineInfo)
        
    def pauseOrRun(self):
        self.pause = not self.pause
        self.runButton.setText("Run" if self.pause else "Pause")
        self.Paused.emit(self.pause)

    def mouseClicked(self, ev):
        if ev.button() == QtCore.Qt.LeftButton:
            self.runButton.clicked.emit()
            #self.pauseOrRun()

    def addData(self, msg):
        # TODO what to do for things that can't be numerically expressed?  just ascii strings, i guess?
        for line in self.lines:
            try:
                if type(msg) != line.msgClass:
                    continue
                newDataPoint = Messaging.getFloat(msg, line.fieldInfo, line.fieldSubindex)
            except ValueError:
                print("ERROR! Plot of %s.%s cannot accept value %s" % (
                    self.msgClass.MsgName(),
                    line.fieldInfo.name,
                    Messaging.get(msg, line.fieldInfo, line.fieldSubindex)))
                continue
            try:
                timestamp = msg.hdr.GetTime()
                if Messaging.findFieldInfo(msg.hdr.fields, "Time").units == "ms":
                    timestamp = timestamp / 1000.0
                newTime = float(elapsedSeconds(timestamp))
                if newTime != 0:
                    self.useHeaderTime = 1
                if not self.useHeaderTime:
                    newTime = elapsedSeconds(datetime.now().timestamp())
            except AttributeError:
                # if header has no time, fallback to PC time.
                newTime = elapsedSeconds(datetime.now().timestamp())
            
            # if the last two data points have the same value as us, then instead of adding a point,
            # we can just change the time of the previous point to our time
            if (len(line.dataArray) > 2 and
                line.dataArray[-2] == line.dataArray[-1] and
                line.dataArray[-1] == newDataPoint):
                line.timeArray[-1] = newTime
            else:
                # add data in the array until MAX_LENGTH is reached, then drop data off start of array
                # such that plot appears to scroll.  The array size is limited to MAX_LENGTH.
                if len(line.dataArray) >= MsgPlot.MAX_LENGTH:
                    line.dataArray[:-1] = line.dataArray[1:]  # shift data in the array one sample left
                    line.dataArray[-1] = newDataPoint
                    line.timeArray[:-1] = line.timeArray[1:]  # shift data in the array one sample left
                    line.timeArray[-1] = newTime
                else:
                    line.dataArray.append(newDataPoint)
                    line.timeArray.append(newTime)

            if not self.pause:
                self.refreshLine(line)

    def refreshLine(self, line):
        timeArray = line.timeArray
        dataArray = line.dataArray
        count = self.timeSlider.value()
        if len(line.dataArray) > count:
            timeArray = timeArray[-count:]
            dataArray = dataArray[-count:]
        line.curve.setData(timeArray, dataArray)
        line.curve.setPos(line.ptr1, 0)
    
    def timeScaleChanged(self):
        for line in self.lines:
            self.refreshLine(line)

    @staticmethod
    def plotFactory(new_plot_callback, msgClass, fieldNames, msgKey = None, fieldLabels = None, runButton = None, clearButton = None, timeSlider = None, displayControls=True):
        msgPlot = None
        if len(fieldNames) == 0:
            fieldNames = [fieldInfo.name for fieldInfo in msgClass.fields]
        idx = 0
        for fieldName in fieldNames:
            fieldLabel = None
            if fieldLabels != None:
                fieldLabel = fieldLabels[idx]
            # if there's a plot widget, try adding a line to it
            if msgPlot != None:
                try:
                    msgPlot.addLine(msgClass, msgKey, fieldName, fieldLabel)
                except MsgPlot.NewPlotError as e:
                    # if error on adding to existing plot, then make a new plot
                    msgPlot = None
                except MsgPlot.PlotError as e:
                    print(str(e))
            
            # make new plot
            if msgPlot == None:
                try:
                    msgPlot = MsgPlot(msgClass, msgKey, fieldName, runButton, clearButton, timeSlider, displayControls, fieldLabel=fieldLabel)
                    msgPlot.msgClass = msgClass
                    new_plot_callback(msgPlot)
                except MsgPlot.PlotError as e:
                    print(str(e))
            idx += 1
        return msgPlot


import msgtools.lib.gui

class MessagePlotGui(msgtools.lib.gui.Gui):
    def __init__(self, argv, parent=None):
        parser = argparse.ArgumentParser(description="Tool to plot message fields")
        parser = msgtools.lib.gui.Gui.addBaseArguments(parser)
        args=parser.parse_args([arg for arg in argv[1:] if not '=' in arg])

        msgtools.lib.gui.Gui.__init__(self, "Message Plot 0.1", args, parent)

        self.plotlayout = QVBoxLayout()
        centralWidget = QWidget()
        centralWidget.setLayout(self.plotlayout)
        self.setCentralWidget(centralWidget)
        self.msgPlots = {}
        self.RxMsg.connect(self.ProcessMessage)

        if len(sys.argv) < 2:
            sys.stderr.write('Usage: ' + sys.argv[0] + ' msg1=field1[,field2] [msg2=field1,field2,field3]\n')
            sys.exit(1)
        
        firstPlot = None
        for arg in argv[1:]:
            if not "=" in arg:
                continue
            argComponentList = arg.split("=")
            if argComponentList[0] == "--log":
                pass
            else:
                msgName = argComponentList[0]
                fieldNameList = argComponentList[1]

                try:
                    msgClass = Messaging.MsgClassFromName[msgName]
                except KeyError:
                    print("ERROR!  Invalid message name " + msgName)
                    continue
                
                if fieldNameList:
                    fieldNames = fieldNameList.split(",")
                else:
                    fieldNames = []
                if firstPlot:
                    MsgPlot.plotFactory(self.newPlot, msgClass, fieldNames, **plotargs)
                else:
                    firstPlot = MsgPlot.plotFactory(self.newPlot, msgClass, fieldNames, displayControls=False)
                    plotargs = {"runButton":firstPlot.runButton, "clearButton":firstPlot.clearButton, "timeSlider":firstPlot.timeSlider, "displayControls":False}
        # add plot controls
        hLayout = QHBoxLayout()
        self.plotlayout.addLayout(hLayout)
        hLayout.addWidget(firstPlot.runButton)
        hLayout.addWidget(firstPlot.clearButton)
        hLayout.addWidget(QLabel("Time Scale"))
        hLayout.addWidget(firstPlot.timeSlider)


    def newPlot(self, plot):
        self.plotlayout.addWidget(QLabel(plot.msgClass.MsgName()))
        self.plotlayout.addWidget(plot)
        if not plot.msgClass.ID in self.msgPlots:
            self.msgPlots[plot.msgClass.ID] = []
        self.msgPlots[plot.msgClass.ID].append(plot)
    
    def ProcessMessage(self, msg):
        if msg.ID in self.msgPlots:
            self.logMsg(msg)
            plotListForID = self.msgPlots[msg.ID]
            for plot in plotListForID:
                plot.addData(msg)

def main(args=None):
    app = QApplication(sys.argv)
    gui = MessagePlotGui(sys.argv)
    gui.show()    
    sys.exit(app.exec_())

# main starts here
if __name__ == '__main__':
    main()



from setuptools import setup, find_packages

setup(name='msgtools',
    python_requires='~=3.5',
    version='0.35.5',
    description='Tools for fixed binary protocols',
    url='https://github.com/MilesEngineering/MsgTools/',
    author='Miles Gazic',
    author_email='miles.gazic@gmail.com',
    license='LGPLv2',
    packages=find_packages(),
    zip_safe=False,
    keywords='development tools messaging messages message generator protocol networking',
    project_urls = {
        'Documentation': 'https://github.com/MilesEngineering/MsgTools/wiki',
        'Source': 'https://github.com/MilesEngineering/MsgTools/',
        'Tracker': 'https://github.com/MilesEngineering/MsgTools/issues',
    },
    classifiers = [
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Embedded Systems',
        'Topic :: Utilities'
    ],
    entry_points = {
        'console_scripts': ['msgparser=msgtools.parser.parser:main',
                            'msgcheck=msgtools.parser.check:main',
                            'msgconsole=msgtools.console.console:main',
                            'msginflux=msgtools.database.influxdb:main',
                            'msginitwebapp=msgtools.webapp.webapp:main',
                            'msgfindprints=msgtools.debug.findprints:main',
                            'msgcsvparse=msgtools.csvparse.csvparse:main'],
        'gui_scripts': ['msgscope=msgtools.scope.scope:main [gui]',
                        'msgserver=msgtools.server.server:main [gui]',
                        'msginspector=msgtools.inspector.inspector:main [gui]',
                        'msgguibuilder=msgtools.guibuilder.guibuilder:main [gui]',
                        'msgscript=msgtools.script.script:main [gui]',
                        'msgplot=msgtools.lib.msgplot:main [gui]',
                        'msgbandwidthtestecho=msgtools.noisemaker.bandwidthtestecho:main [gui]',
                        'msgbandwidthtester=msgtools.noisemaker.bandwidthtester:main [gui]',
                        'msgnoisemaker=msgtools.noisemaker.noisemaker:main [gui]',
                        'msglumberjack=msgtools.lumberjack.lumberjack:main [gui]',
                        'msgdebug=msgtools.debug.debug:main [gui]',
                        'msglauncher=msgtools.launcher.launcher:main [gui]'],
        'msgtools.parser.plugin': ['c=msgtools.parser.c.language',
                                   'cpp=msgtools.parser.cpp.language',
                                   'java=msgtools.parser.java.language',
                                   'javascript=msgtools.parser.javascript.language',
                                   'python=msgtools.parser.python.language',
                                   'html=msgtools.parser.html.language',
                                   'matlab=msgtools.parser.matlab.language',
                                   'simulink=msgtools.parser.simulink.language',
                                   'swift=msgtools.parser.swift.language', 
                                   'kotlin=msgtools.parser.kotlin.language'],
        'msgtools.server.plugin': ['serial=msgtools.server.SerialPlugin:plugin_info',
                                   'bluetoothSPP=msgtools.server.SerialPlugin:bt_plugin_info',
                                   'bluetoothRFCOMM=msgtools.server.BluetoothRFCOMM:plugin_info',
                                   'bluetoothRFCOMMQt=msgtools.server.BluetoothRFCOMMQt:plugin_info',
                                   'influxdb=msgtools.database.influx_msgserver_plugin:plugin_info'],
        'msgtools.launcher.plugin': ['scope=msgtools.scope.launcher:info',
                                   'script=msgtools.script.launcher:info',
                                   'server=msgtools.server.launcher:info',
                                   'inspector=msgtools.inspector.launcher:info',
                                   'debug=msgtools.debug.launcher:info'],
    },
    install_requires=[
        'pyyaml',
        'websockets',
        'janus',
        'jinja2',
        'numpy'
    ],
    extras_require={
        'gui':  ["pyqtgraph", "pyqt5", 'qscintilla'],
        'serverserial': ["pyqtserial"],
    },
    package_data={
        # Include all Template files for the code generator and web app tool
        '': ['*Template*', 'bootstrap.min.css', 'webapp/template.*', 'webapp/lib/**', '*.png']
    }
)



#!/cygdrive/c/Python34/python.exe
import sys

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from msgtools.lib.messaging import Messaging

class FieldItem(QTreeWidgetItem):
    def __init__(self, msg, msg_key, fieldInfo, column_strings = []):
        column_strings = column_strings if len(column_strings) > 0 else [None, fieldInfo.name, "", fieldInfo.units, fieldInfo.description]
        
        QTreeWidgetItem.__init__(self, None, column_strings)
        
        self.fieldInfo = fieldInfo
        self.fieldName = fieldInfo.name
        self.msg = msg
        self.msg_key = msg_key
        
    def data(self, column, role):
        if column != 2:
            return super(FieldItem, self).data(column, role)

        alert = Messaging.getAlert(self.msg, self.fieldInfo)
        if role == Qt.FontRole:
            font = QFont()
            if alert == 1:
                font.setBold(1)
            return font
        if role == Qt.ForegroundRole:
            brush = QBrush()
            if alert == 1:
                brush.setColor(Qt.red)
            return brush

        if role == Qt.DisplayRole:
            value  = Messaging.get(self.msg, self.fieldInfo)
            
            try:
                self.overrideWidget
                valueAsString = Messaging.get(self.msg, self.fieldInfo)
                
                if valueAsString in self.comboBoxIndexOfEnum:
                    #self.overrideWidget.setCurrentText(valueAsString)
                    self.overrideWidget.setCurrentIndex(self.comboBoxIndexOfEnum[valueAsString])
                else:
                    #self.overrideWidget.setEditText(valueAsString)
                    self.overrideWidget.setCurrentIndex(-1)
            except AttributeError:
                pass
                
            return str(value)
            
        return super(FieldItem, self).data(column, role)

class EditableFieldItem(FieldItem):
    def __init__(self, msg, msg_key, fieldInfo, column_strings = []):
        super(EditableFieldItem, self).__init__(msg, msg_key, fieldInfo, column_strings)

        self.setFlags(self.flags() | Qt.ItemIsEditable)
        
        if fieldInfo.type == "enumeration":
            self.overrideWidget = QComboBox()
            self.overrideWidget.addItems(fieldInfo.enum[0].keys())
            self.overrideWidget.activated.connect(self.overrideWidgetValueChanged)
            # there's some odd behavior in the UI when the box is editable :(
            # if you want it editable, uncomment this line, and play around and see if you like it
            #self.overrideWidget.setEditable(1)
            # store a hash table of value->ComboBoxIndex
            # this is NOT the same as value->enumIndex!
            self.comboBoxIndexOfEnum = {}
            for i in range(0, self.overrideWidget.count()):
                self.comboBoxIndexOfEnum[self.overrideWidget.itemText(i)] = i

    def overrideWidgetValueChanged(self, value):
        valueAsString = self.overrideWidget.itemText(value)
        # set the value in the message/header buffer
        Messaging.set(self.msg, self.fieldInfo, valueAsString)

        # no need to reset UI to value read from message, if user picked value from drop down.
        # \todo: need to if they type something, though.
        
    def setData(self, column, role, value):
        if not column == 2:
            return

        if self.fieldInfo.name == "ID":
            return
        
        if self.fieldInfo.type == "int" and value.startswith("0x"):
            value = str(int(value, 0))

        # if user deletes the value, for anything besides a string,
        # return without setting the new value
        if self.fieldInfo.type != "string" and value == "":
            return

        # set the value in the message/header buffer
        Messaging.set(self.msg, self.fieldInfo, value)

        # get the value back from the message/header buffer and pass on to super-class' setData
        super(FieldItem, self).setData(column, role, Messaging.get(self.msg, self.fieldInfo))

class FieldBits(FieldItem):
    def __init__(self, msg, msg_key, bitfieldInfo):
       column_strings = [None, "    " + bitfieldInfo.name, "", bitfieldInfo.units, bitfieldInfo.description]
       super(FieldBits, self).__init__(msg, msg_key, bitfieldInfo, column_strings)

class FieldBitfieldItem(FieldItem):
    def __init__(self, tree_widget, msg, msg_key, fieldInfo):
        super(FieldBitfieldItem, self).__init__(msg, msg_key, fieldInfo)

        for bitfieldInfo in fieldInfo.bitfieldInfo:
            bitfieldBitsItem = FieldBits(self.msg, msg_key, bitfieldInfo)
            self.addChild(bitfieldBitsItem)

class EditableFieldBits(EditableFieldItem):
    def __init__(self, msg, msg_key, bitfieldInfo):
       column_strings = [None, "    " + bitfieldInfo.name, "", bitfieldInfo.units, bitfieldInfo.description]
       super(EditableFieldBits, self).__init__(msg, msg_key, bitfieldInfo, column_strings)

class EditableFieldBitfieldItem(EditableFieldItem):
    def __init__(self, tree_widget, msg, msg_key, fieldInfo):
        super(EditableFieldBitfieldItem, self).__init__(msg, msg_key, fieldInfo)

        for bitfieldInfo in fieldInfo.bitfieldInfo:
            bitfieldBitsItem = EditableFieldBits(self.msg, msg_key, bitfieldInfo)
            self.addChild(bitfieldBitsItem)
            try:
                bitfieldBitsItem.overrideWidget
                tree_widget.setItemWidget(bitfieldBitsItem, 2, bitfieldBitsItem.overrideWidget)
            except AttributeError:
                pass

class FieldArrayItem(QTreeWidgetItem):
    def __init__(self, msg, msg_key, fieldInfo, field_array_constructor, index = None):
        column_strings = [None, fieldInfo.name, "", fieldInfo.units, fieldInfo.description]
        
        if index != None:
            column_strings[1] = "    [" + str(index) + "]"
        
        QTreeWidgetItem.__init__(self, None, column_strings)
        
        self.fieldInfo = fieldInfo
        if index == None:
            self.fieldName = fieldInfo.name
        else:
            self.fieldName = "%s[%d]" % (fieldInfo.name, index)
        self.msg = msg
        self.msg_key = msg_key
        self.index = index

        if index == None:
            for i in range(0, self.fieldInfo.count):
                messageFieldTreeItem = field_array_constructor(self.msg, self.msg_key, self.fieldInfo, field_array_constructor, i)
                self.addChild(messageFieldTreeItem)

    def data(self, column, role):
        if column != 2:
            return super(FieldArrayItem, self).data(column, role)

        if self.index == None:
            if role == Qt.FontRole:
                return QFont()
            return ""

        alert = Messaging.getAlert(self.msg, self.fieldInfo, self.index)
        if role == Qt.FontRole:
            font = QFont()
            if alert == 1:
                font.setBold(1)
            return font
        if role == Qt.ForegroundRole:
            brush = QBrush()
            if alert == 1:
                brush.setColor(Qt.red)
            return brush

        if role == Qt.DisplayRole:
            value  = Messaging.get(self.msg, self.fieldInfo, self.index)
            return str(value)

        return super(FieldArrayItem, self).data(column, role)

class EditableFieldArrayItem(FieldArrayItem):
    def __init__(self, msg, msg_key, fieldInfo, field_array_constructor, index = None):
        super(EditableFieldArrayItem, self).__init__(msg, msg_key, fieldInfo, field_array_constructor, index)

        self.setFlags(self.flags() | Qt.ItemIsEditable)

    def setData(self, column, role, value):
        if self.index == None:
            return

        if column != 2:
            return

        if self.fieldInfo.name == "ID":
            return

        if self.fieldInfo.type == "int" and value.startswith("0x"):
            value = str(int(value, 0))

        # set the value in the message/header buffer
        Messaging.set(self.msg, self.fieldInfo, value, int(self.index))

        # get the value back from the message/header buffer and pass on to super-class' setData
        super(EditableFieldArrayItem, self).setData(column, role, Messaging.get(self.msg, self.fieldInfo, int(self.index)))

class QObjectProxy(QObject):
    send_message = pyqtSignal(object)
    def __init__(self):
        QObject.__init__(self)

class MessageItem(QTreeWidgetItem):
    def __init__(self, tree_widget, msg, msg_key,
                 child_constructor = FieldItem,
                 child_array_constructor = FieldArrayItem,
                 child_bitfield_constructor = FieldBitfieldItem):
        QTreeWidgetItem.__init__(self, None, [msg.MsgName()])
        
        self.qobjectProxy = QObjectProxy()

        self.tree_widget = tree_widget

        self.msg = msg
        self.msg_key = msg_key

        self.setup_fields(tree_widget, child_constructor, child_array_constructor, child_bitfield_constructor)

        tree_widget.addTopLevelItem(self)
        tree_widget.resizeColumnToContents(0)
        self.setExpanded(True)

    def repaintAll(self):
        # Refresh the paint on the entire tree
        # TODO This is not a good solution!  We should refresh *only* the item that changed, not whole tree!
        region = self.tree_widget.childrenRegion()
        self.tree_widget.setDirtyRegion(region)

    def set_msg_buffer(self, msg_buffer):
        self.msg.msg_buffer_wrapper["msg_buffer"] = msg_buffer
        self.msg.hdr.msg_buffer_wrapper["msg_buffer"] = msg_buffer
        self.repaintAll()

    def setup_fields(self, tree_widget, child_constructor, child_array_constructor, child_bitfield_constructor):
        headerTreeItemParent = QTreeWidgetItem(None, [ "Header" ])
        self.addChild(headerTreeItemParent)

        for headerFieldInfo in Messaging.hdr.fields:
            if headerFieldInfo.bitfieldInfo != None:
                headerFieldTreeItem = child_bitfield_constructor(tree_widget, self.msg.hdr, None, headerFieldInfo)
            else:
                headerFieldTreeItem = child_constructor(self.msg.hdr, None, headerFieldInfo)
            headerTreeItemParent.addChild(headerFieldTreeItem)

        for fieldInfo in type(self.msg).fields:
            messageFieldTreeItem = None

            if fieldInfo.count == 1:
                if fieldInfo.bitfieldInfo != None:
                    messageFieldTreeItem = child_bitfield_constructor(tree_widget, self.msg, self.msg_key, fieldInfo)
                else:
                    messageFieldTreeItem = child_constructor(self.msg, self.msg_key, fieldInfo)
            else:
                messageFieldTreeItem = child_array_constructor(self.msg, self.msg_key, fieldInfo, child_array_constructor)
            
            self.addChild(messageFieldTreeItem)
            try:
                messageFieldTreeItem.overrideWidget
                tree_widget.setItemWidget(messageFieldTreeItem, 2, messageFieldTreeItem.overrideWidget)
            except AttributeError:
                pass

class EditableMessageItem(MessageItem):

    def __init__(self, tree_widget, msg, msg_key):
        super(EditableMessageItem, self).__init__(tree_widget, msg, msg_key, EditableFieldItem, EditableFieldArrayItem, EditableFieldBitfieldItem)

        # text entry for rate, and a 'send' button
        self.sendTimer = QTimer()
        self.sendTimer.timeout.connect(lambda: self.qobjectProxy.send_message.emit(self.msg))
        containerWidget = QWidget(tree_widget)
        containerLayout = QHBoxLayout()
        containerWidget.setLayout(containerLayout)
        self.sendButton = QPushButton("Send", tree_widget)
        self.sendButton.autoFillBackground()
        self.sendButton.clicked.connect(self.sendClicked)
        self.rateEdit = QLineEdit("", tree_widget)
        self.rateEdit.setValidator(QDoubleValidator(0, 100, 2, containerWidget))
        self.rateEdit.setMaximumWidth(50)
        self.rateEdit.setFixedWidth(50)
        self.rateEdit.setPlaceholderText("Rate")
        containerLayout.addWidget(self.rateEdit)
        containerLayout.addWidget(QLabel("Hz"))
        containerLayout.addWidget(self.sendButton)
        tree_widget.setItemWidget(self, 4, containerWidget)
        
        for i in range(0, tree_widget.columnCount()):
            tree_widget.resizeColumnToContents(i);

    def sendClicked(self):
        if self.sendButton.text() == "Send":
            if self.rateEdit.text() == "" or self.rateEdit.text() == "0":
                self.qobjectProxy.send_message.emit(self.msg)
            else:
                self.sendButton.setText("Stop")
                rate = float(self.rateEdit.text())
                if rate > 0.01:
                    period = 1000.0/rate
                    # send once now, then periodically after
                    self.qobjectProxy.send_message.emit(self.msg)
                    self.sendTimer.start(period)
                else:
                    self.sendButton.text("Send")
        else:
            self.sendButton.setText("Send")
            self.sendTimer.stop()
